# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
import os
from pathlib import Path
from typing import Union, List
from shutil import copy2, copytree, rmtree
try:
    import importlib.resources as importlib_resources
except ModuleNotFoundError:
    import importlib_resources

from towhee.utils.singleton import singleton
from towhee.engine import DEFAULT_LOCAL_CACHE_ROOT
from towhee.utils.log import engine_log
from towhee.hub.pipeline_manager import PipelineManager
from towhee.hub.operator_manager import OperatorManager
from towhee.utils.git_utils import GitUtils


@singleton
class FileManagerConfig():
    """
    Global FileManager config

    This class dictates which cache locations to look through and in what order to check
    for files. This config is ultimately used with FileManager.

    Args:
        set_default_cache (`str` | `Path`):
            The default cache to check in, if nothing supplied, the default cache of $HOME/.towhee
            will be used.
    """
    def __init__(self):
        # TODO: #1 Deal with specifying cache priority per pipeline?
        self._cache_paths = [DEFAULT_LOCAL_CACHE_ROOT]
        self._cache_operators = []
        self._cache_pipelines = []

        self._cache_paths_lock = threading.Lock()
        self._operators_lock = threading.Lock()
        self._pipelines_lock = threading.Lock()

    # Requires manual locking
    @property
    def default_cache(self):
        return self._cache_paths[-1]

    # Requires manual locking
    @property
    def cache_paths(self):
        return self._cache_paths

    # Requires manual locking
    @property
    def cache_operators(self):
        return self._cache_operators

    # Requires manual locking
    @property
    def cache_pipelines(self):
        return self._cache_pipelines

    @property
    def operator_lock(self):
        return self._operators_lock

    @property
    def pipeline_lock(self):
        return self._pipelines_lock

    @property
    def cache_paths_lock(self):
        return self._cache_paths_lock

    def update_default_cache(self, default_path: Union[str, Path]):
        """
        Change the default cache path.

        Args:
            insert_path (`str` | `Path`):
                The new default cache.
        """
        default_path = Path(default_path)
        with self._cache_paths_lock:
            self._cache_paths.append(default_path)

    def add_cache_path(self, insert_path: Union[str, Path, List[Union[str, Path]]]):
        """
        Add a cache location to the front. Most recently added paths will be
        checked first.

        Args:
            insert_path (`str` | `Path | `list[str | Path]`):
                The path that you are trying to add. Accepts multiple inputs at once.
        """
        with self._cache_paths_lock:
            if not isinstance(insert_path, list):
                insert_path = [insert_path]

            for path in insert_path:
                if str(path) not in [str(x) for x in self._cache_paths]:
                    self._cache_paths.insert(0, Path(path))

    # Pretty useless now but might be worth having in the future
    def remove_cache_path(self, remove_path: Union[str, Path, List[Union[str, Path]]]):
        """
        Remove a cache location.

        Args:
            remove_path (`str` | `Path | `list[str | Path]`):
                The path that you are trying to remove. Accepts multiple inputs at once.
        """
        if not isinstance(remove_path, list):
            remove_path = [remove_path]

        with self._cache_paths_lock:
            default_path = self._cache_paths[-1]
            for path in remove_path:
                try:
                    self._cache_paths.remove(Path(path))
                except ValueError:
                    engine_log.error('%s%s ', str(Path(path)), ' not found.')

            if len(self._cache_paths) <= 1:
                self._cache_paths.append(default_path)

    def reset_cache_path(self):
        """
        Remove all cache locations.

        """
        with self._cache_paths_lock:
            self._cache_paths = [self._cache_paths.pop()]

    def cache_local_operator(self, path: Union[str, Path, List[Union[str, Path]]], overwrite: bool = True, cache: Union[str, Path] = None):
        """
        Add a local operator to a cache. In order for operators to be visible to Towhee
        they need to be cached.

        Args:
            path (`str` | `Path | `list[str | Path]`):
                The path to the operator or its directory.
            overwrite (`bool`):
                Whether to overwrite already existing operator in cache.
            cache (`str` | `Path`):
                Optional location to cache the opeartor.
        """
        if not isinstance(path, list):
            path = [path]

        with self._operators_lock:
            if cache is None:
                with self._cache_paths_lock:
                    cache = self._cache_paths[-1]
            for paths in path:
                op = {}
                op['path'] = Path(paths)
                op['name'] = op['path'].name
                op['cache'] = Path(cache)
                op['overwrite'] = overwrite

                if op not in self._cache_operators:
                    self._cache_operators.append(op)

    def cache_local_pipeline(self, path: Union[str, Path, List[Union[str, Path]]], overwrite: bool = True, cache: Union[str, Path] = None):
        """
        Add a local pipeline to a cache. In order for pipelines to be visible to Towhee
        they need to be cached.

        Args:
            path (`str` | `Path | `list[str | Path]`):
                The path to the pipeline or its directory.
            overwrite (`bool`):
                Whether to overwrite already existing operator in cache.
            cache (`str` | `Path`):
                Optional location to cache the opeartor.
        """
        if not isinstance(path, list):
            path = [path]

        with self._pipelines_lock:
            if cache is None:
                with self._cache_paths_lock:
                    cache = self._cache_paths[-1]
            for paths in path:
                op = {}
                op['path'] = Path(paths)
                op['name'] = op['path'].name
                op['cache'] = Path(cache)
                op['overwrite'] = overwrite

                if op not in self._cache_pipelines:
                    self._cache_pipelines.append(op)

    def __str__(self):
        return str(self.__dict__)


@singleton
class FileManager():
    """
    The filemanager for the current instance of Towhee. In charge of dealing with multiple
    cache locations, downloading from hub, and importing local files.

    Args:
        fmc (`FileManagerConfig`):
            Accepts an optional FileManager config, once a FileManagerConfig is selected,
            it cannot be changed for the current runtime.
    """
    def __init__(self, fmc: FileManagerConfig = FileManagerConfig()):
        self._config = fmc
        # TODO: #1 seperate ranking for different pipelines?
        # self._pipelines = []
        self._operator_lock = threading.Lock()
        self._pipeline_lock = threading.Lock()
        self._cache_locals()

    # Move to a utils location?
    def _cache_name(self, repo: str, author: str, tag: str):
        try:
            file_name, file_type = repo.split('.')
        except ValueError:
            return Path(author) / repo.replace('-', '_') / tag
        # If pointing directly to pipeline or operator file.
        if file_type in ['py', 'yaml']:
            return Path(author) / file_name.replace('-', '_') / tag
        # If not pointing to a pipeline or an operator file.
        else:
            raise ValueError('Unsupported file type.')

    # TODO: filip-halt
    # Decide if to treat pipelines and operators the same. Should we allow single .py
    # for opeartor and should we allow directory for pipeline.
    def _cache_locals(self):
        with self._pipeline_lock and self._config.pipeline_lock:
            for pipe in self._config.cache_pipelines:
                new_dir = self._cache_name(pipe['name'], author='local', tag='main')
                file = pipe['name']
                cache_path = pipe['cache']
                old_path = pipe['path']
                new_path = cache_path / 'pipelines' / new_dir / file

                if pipe['overwrite']:
                    Path.unlink(new_path, missing_ok=True)

                if new_path.is_file():
                    # TODO: filip-halt
                    # Error logging.
                    engine_log.info('Did not overwrite.')
                else:
                    # TODO: filip-halt
                    # Figure out exception types.
                    new_path.parent.mkdir(parents=True, exist_ok=True)
                    copy2(str(old_path), str(new_path))

        with self._operator_lock and self._config.operator_lock:
            for op in self._config.cache_operators:
                new_dir = self._cache_name(op['name'], author='local', tag='main')
                cache_path = op['cache']
                old_path = op['path']
                new_path = cache_path / 'operators' / new_dir

                if op['overwrite'] and new_path.is_dir():
                    rmtree(new_path)

                if new_path.is_dir():
                    # TODO: filip-halt
                    # Error logging.
                    engine_log('Did not overwrite.')
                else:
                    # TODO: Figure out exception types
                    copytree(str(old_path), str(new_path))

    def get_builtin_pipeline(self, file_name: str) -> str:
        with importlib_resources.path('towhee.hub.builtin.pipelines', '{}.yaml'.format(file_name)) as path:
            return path

    def get_pipeline(self, pipeline: str, tag: str, install_reqs: bool = True):
        """
        Obtain the path to the requested pipeline.

        This function will obtain the first reference to the pipeline from the cache locations.
        If no pipeline is found, this function will download it to the default cache location
        from the Towhee hub.

        Args:
            pipeline (`str`):
                The pipeline in 'author/repo' format. Author will be 'local' if locally imported.
            tag (`str`):
                Which tag to use of the pipeline. Will use 'main' if locally imported.
            install_reqs (`bool`):
                Whether to download the python packages if a requirements.txt file is included in the repo.

        Returns:
            (Path | None)
                Returns the path of the pipeline, None if a local pipeline isnt found.

        Raises:
            (`ValueError`):
                Incorrect pipeline format.
        """
        with self._pipeline_lock:
            pipeline_split = pipeline.split('/')
            # For now assuming all piplines will be classifed as 'author/repo'.
            if len(pipeline_split) != 2:
                raise ValueError('''Incorrect pipeline format, should be '<author>/<pipeline_repo>'.''')
            author = pipeline_split[0]
            repo = pipeline_split[1]
            file_name = repo.replace('-', '_')

            if author == 'builtin':
                return self.get_builtin_pipeline(file_name)

            # This path leads to 'author/repo/tag/file_name.yaml'
            pipeline_path = self._cache_name(repo, author, tag) / (file_name + '.yaml')

            found_existing = False
            for path_iter in self._config.cache_paths:
                # This path leads to 'cache/pipelines/author/repo/tag/file_name.yaml'
                path = path_iter / 'pipelines' / pipeline_path
                if path.is_file():
                    file_path = path
                    found_existing = True
                    break

            if author == 'local':
                if found_existing is False:
                    msg = f'Pipeline {pipeline} not find, has it been imported?'
                    raise FileNotFoundError(msg)
                return file_path

            if not found_existing:
                file_path = self._config.default_cache / 'pipelines' / pipeline_path
                repo_path = file_path.parent

                # If yaml file missing but repo exists.
                if repo_path.is_dir():
                    engine_log.warning('Your local opreator repo %s exists, but not complete, \'.py\' file missing, please check on hub.', file_name)
                    rmtree(repo_path)
                    raise FileNotFoundError('Repo exists but \'.yaml\' file missing.')

                # If user has git, clone the repo, otherwise try to download.
                try:
                    git = GitUtils(author, repo)
                    git.clone(tag=tag, install_reqs=install_reqs, local_repo_path=repo_path)
                # If user does not have git in the system, subprocess throws FileNotFound error.
                except FileNotFoundError:
                    pipeline_manager = PipelineManager(author=author, repo=repo)
                    pipeline_manager.download(local_repo_path=file_path.parent, tag=tag, install_reqs=install_reqs)
                    return file_path

            # Check updates for repo.
            elif tag == 'main':
                repo_path = file_path.parent
                cwd = Path.cwd()
                os.chdir(repo_path)
                try:
                    git = GitUtils(author, repo)
                    if 'Your branch is behind' in git.status():
                        engine_log.warning('Your local pipeline %s is not up to date, updating to latest version...', file_name)
                        git.pull()
                except FileNotFoundError:
                    pass
                os.chdir(cwd)

        return file_path

    def get_operator(self, operator: str, tag, install_reqs: bool = True):
        """
        Obtain the path to the requested operator.

        This function will obtain the first reference to the operator from the cache locations.
        If no opeartor is found, this function will download it to the default cache location
        from the Towhee hub.

        Args:
            operator (`str`):
                The operator in 'author/repo' format. Author will be 'local' if locally imported.
            tag (`str`):
                Which tag version to use of the opeartor. Will use 'main' if locally imported.
            install_reqs (`bool`):
                Whether to download the python packages if a requirements.txt file is included in the repo.

        Returns:
            (Path | None)
                Returns the path of the operator, None if a local operator isnt found.

        Raises:
            (`ValueError`):
                Incorrect opeartor format.
        """
        with self._operator_lock:
            operator_split = operator.split('/')
            # For now assuming all piplines will be classifed as 'author/repo'.
            if len(operator_split) != 2:
                raise ValueError('''Incorrect operator format, should be '<author>/<operator_repo>'.''')
            author = operator_split[0]
            repo = operator_split[1]
            file_name = repo.replace('-', '_')
            # This path leads to 'author/repo/tag/file_name.yaml'
            operator_path = self._cache_name(repo, author, tag) / (file_name + '.py')
            init_path = self._cache_name(repo, author, tag) / ('__init__.py')

            found_existing = False
            for path_iter in self._config.cache_paths:
                # This path leads to 'cache/operators/author/repo/tag/file_name.py'
                path = path_iter / 'operators' / operator_path
                path2 = path_iter / 'operators' / init_path
                if path.is_file() or path2.is_file():
                    file_path = path
                    found_existing = True
                    break

            if author == 'local':
                if found_existing is False:
                    msg = f'Operator {operator} not found, has it been imported?'
                    raise FileNotFoundError(msg)
                return file_path

            if not found_existing:
                file_path = self._config.default_cache / 'operators' / operator_path
                repo_path = file_path.parent
                # If py file missing but repo exists.
                if repo_path.is_dir():
                    engine_log.error('Your local opreator repo %s exists, but not complete, \'.py\' file missing, please check on hub.', file_name)
                    rmtree(repo_path)
                    raise FileNotFoundError('Repo exists but \'.py\' file missing.')

                # If user has git, clone the repo, otherwise try to download.
                try:
                    git = GitUtils(author, repo)
                    git.clone(tag=tag, install_reqs=install_reqs, local_repo_path=repo_path)
                # If user does not have git in the system, subprocess throws FileNotFound error.
                except FileNotFoundError:
                    operator_manager = OperatorManager(author=author, repo=repo)
                    operator_manager.download(local_repo_path=file_path.parent, tag=tag, install_reqs=install_reqs)
                    return file_path

            # Check updates fro repo.
            elif tag == 'main':
                repo_path = file_path.parent
                cwd = Path.cwd()
                os.chdir(repo_path)
                try:
                    git = GitUtils(author, repo)
                    if 'Your branch is behind' in git.status():
                        engine_log.warning('Your local operator %s is not up to date, updating to latest version...', file_name)
                        git.pull()
                except FileNotFoundError:
                    pass
                os.chdir(cwd)

        return file_path



from torch.utils import model_zoo
import torch
import logging

def load_pretrained_weights(
    model,
    model_name=None,
    model_configs=None,
    weights_path=None,
):
    """
    Load pretrained weights from path or url.
    Args:
        model(nn.Module):
            Model.
        model_name(str):
            Model name
        model_configs(dict):
            Model configs
        weights_path(str, optional):
            Path to pretrained weights file on the local disk.
        verbose (bool):
            If printing is done when downloading is over.
    """
    logging.info('model_name is %s', str(model_name))
    logging.info('weights_path is %s', str(weights_path))
    assert bool(model_name), 'Expected model_name'

    # Load or download weights
    if weights_path is None:
        logging.info(model_configs.keys())
        url = model_configs['url']
        if url:
            logging.info('Please check hub connection in case weights can not be downloaded!')
            state_dict = model_zoo.load_url(url, map_location=torch.device('cpu'))
        else:
            raise ValueError(f'Pretrained model for {model_name} is not available.')
    else:
        state_dict = torch.load(weights_path)

   # Load state dict
    if 'model' in state_dict:
        ret = model.load_state_dict(state_dict['model'], strict=False)
    else:
        ret = model.load_state_dict(state_dict, strict=False)
    return ret



# Original implementation by Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.
#
# Modifications by Copyright 2022 Zilliz . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os

import torch
from torch import nn

from torchvision.transforms import Compose, Lambda

try:
    from torchvideo.transforms import (
        CenterCropVideo,
        NormalizeVideo,
        CollectFrames,
        # PILVideoToTensor
    )
except ModuleNotFoundError:
    os.system('pip install "git+https://github.com/willprice/torchvideo.git"')
    from torchvideo.transforms import (
        CenterCropVideo,
        NormalizeVideo,
        CollectFrames,
        # PILVideoToTensor
    )

try:
    from pytorchvideo.data.encoded_video import EncodedVideo
except ModuleNotFoundError:
    os.system('pip install "git+https://github.com/facebookresearch/pytorchvideo.git"')
    from pytorchvideo.data.encoded_video import EncodedVideo
from pytorchvideo.transforms import (
    ApplyTransformToKey,
    ShortSideScale,
    UniformTemporalSubsample,
    # UniformCropVideo
)


class VideoTransforms:
    """
    Transform video to video & audio tensors given an video path, corresponding to models by model name.
    The code is buiil on top of examples given by [Pytorchvideo](https://pytorchvideo.org/).
    The following model names are supported:
        - slow_r50
        - slowfast_r50
        - slowfast_r101
        - x3d_xs
        - x3d_s
        - x3d_m

    Args:
        model_name (`str`):
            model name
    Returns:
        A dictionary including tensors for both video and audio.

    Example:
        >>> from towhee.models.utils.video_transforms import VideoTransforms
        >>> tfms = VideoTransforms("slowfast_r50")
        >>> video_data = tfms(video_path="path/to/video", start_sec=0, end_sec=30)
    """

    def __init__(self, model_name: str):
        self.model_name = model_name
        self.tfms_params = video_configs[model_name]
        self.tfms = ApplyTransformToKey(
            key="video",
            transform=Compose([
                UniformTemporalSubsample(self.tfms_params["num_frames"]),
                Lambda(lambda x: x / 255.0),
                NormalizeVideo(
                    mean=self.tfms_params["mean"],
                    std=self.tfms_params["std"],
                    inplace=True
                ),
                ShortSideScale(size=self.tfms_params["side_size"]),
                CenterCropVideo(
                    size=(self.tfms_params["crop_size"], self.tfms_params["crop_size"])
                ),
                CollectFrames(),
                PackPathway(alpha=self.tfms_params["alpha"]) if model_name.startswith("slowfast") else nn.Identity()
            ]),
        )

    def __call__(self, video_path: str, start_sec: float = 0.0, end_sec: float = 30.0):
        video = EncodedVideo.from_path(video_path)
        video_data = video.get_clip(start_sec=start_sec, end_sec=end_sec)
        video_data = self.tfms(video_data)
        if isinstance(video_data["video"], list) and str(self.tfms._transform.transforms[-1]) == "PackPathway()":
            video_data["video_slow"] = video_data["video"][0]
            video_data["video_fast"] = video_data["video"][1]
        else:
            video_data["video"] = torch.stack(video_data["video"])
        return video_data


class PackPathway(nn.Module):
    """
    Transform for converting video frames as a list of tensors.

    Args:
        alpha (`int`):
            alpha value

    Returns:
        a list of tensors [slow_pathway, fast_pathway]
    """

    def __init__(self, alpha):
        super().__init__()
        self.alpha = alpha

    def forward(self, frames: torch.Tensor):
        frames = torch.stack(frames)
        fast_pathway = frames
        # Perform temporal sampling from the fast pathway.
        slow_pathway = torch.index_select(
            frames,
            1,
            torch.linspace(
                0, frames.shape[1] - 1, frames.shape[1] // self.alpha
            ).long(),
        )
        frame_list = [slow_pathway, fast_pathway]
        return frame_list


video_configs = {
    "slow_r50": {
        "side_size": 256,
        "crop_size": 256,
        "num_frames": 8,
        "sampling_rate": 8,
        "mean": [0.45, 0.45, 0.45],
        "std": [0.225, 0.225, 0.225],
    },
    "slowfast_r50": {
        "side_size": 256,
        "crop_size": 256,
        "num_frames": 32,
        "sampling_rate": 2,
        "alpha": 4,
        "mean": [0.45, 0.45, 0.45],
        "std": [0.225, 0.225, 0.225],
    },
    "slowfast_r101": {
        "side_size": 256,
        "crop_size": 256,
        "num_frames": 32,
        "sampling_rate": 8,
        "alpha": 4,
        "mean": [0.45, 0.45, 0.45],
        "std": [0.225, 0.225, 0.225],
    },
    "x3d_xs": {
        "side_size": 182,
        "crop_size": 182,
        "num_frames": 4,
        "sampling_rate": 12,
        "mean": [0.45, 0.45, 0.45],
        "std": [0.225, 0.225, 0.225],
    },
    "x3d_s": {
        "side_size": 182,
        "crop_size": 182,
        "num_frames": 13,
        "sampling_rate": 6,
        "mean": [0.45, 0.45, 0.45],
        "std": [0.225, 0.225, 0.225],
    },
    "x3d_m": {
        "side_size": 256,
        "crop_size": 256,
        "num_frames": 16,
        "sampling_rate": 5,
        "mean": [0.45, 0.45, 0.45],
        "std": [0.225, 0.225, 0.225],
    },
    "mvit_base_16x4": {
            "side_size": 224,
            "crop_size": 224,
            "num_frames": 16,
            "sampling_rate": 4,
            "mean": [0.45, 0.45, 0.45],
            "std": [0.225, 0.225, 0.225],
        },
    "mvit_base_32x3": {
            "side_size": 224,
            "crop_size": 224,
            "num_frames": 32,
            "sampling_rate": 3,
            "mean": [0.45, 0.45, 0.45],
            "std": [0.225, 0.225, 0.225],
        },
}



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
import math
from typing import Callable, Tuple
from functools import partial

import numpy as np
import torch
from torch import nn
from torch.nn import functional as F

from towhee.models.utils.general_utils import to_2tuple
from towhee.models.layers.conv2d_same import conv2d_same
from towhee.models.layers.padding_functions import get_padding_value

def get_condconv_initializer(initializer: Callable, num_experts: int, expert_shape: Tuple[int, int, int, int]):
    def condconv_initializer(weight):
        """CondConv initializer function."""
        num_params = np.prod(expert_shape)
        if (len(weight.shape) != 2 or weight.shape[0] != num_experts or
                weight.shape[1] != num_params):
            raise (ValueError(
                'CondConv variables must have shape [num_experts, num_params]'))
        for i in range(num_experts):
            initializer(weight[i].view(expert_shape))
    return condconv_initializer

class CondConv2d(nn.Module):
    """
    Conditionally Parameterized Convolution

    Inspired by: https://github.com/tensorflow/tpu/blob/master/models/official/efficientnet/condconv/condconv_layers.py
    Grouped convolution hackery for parallel execution of the per-sample kernel filters inspired by this discussion:
    https://github.com/pytorch/pytorch/issues/17983

    Args:
        in_channels (`int`):
            Number of channels in the input image
        out_channels (`int`):
            Number of channels produced by the convolution.
        kernel_size (`int`):
            Size of the convolving kernel.
        stride (`int`):
            Stride of the convolution.
        padding (`str`):
            Padding added to all four sides of the input.
        dilation (`int`):
            Spacing between kernel elements.
        groups (`int`):
            Number of blocked connections from input channels to output channels.
        bias (`bool`):
            If True, adds a learnable bias to the output.
        num_experts (`int`):
            The number of expert kernels and biases.
    """
    __constants__ = ['in_channels', 'out_channels', 'dynamic_padding']

    def __init__(self, in_channels: int, out_channels: int, kernel_size: int=3,
                 stride: int=1, padding: str='', dilation: int=1, groups: int=1, bias: bool=False, num_experts: int=4) -> None:
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = to_2tuple(kernel_size)
        self.stride = to_2tuple(stride)
        padding_val, is_padding_dynamic = get_padding_value(
            padding, kernel_size, stride=stride, dilation=dilation)
        self.dynamic_padding = is_padding_dynamic  # if in forward to work with torchscript
        self.padding = to_2tuple(padding_val)
        self.dilation = to_2tuple(dilation)
        self.groups = groups
        self.num_experts = num_experts

        self.weight_shape = (self.out_channels, self.in_channels // self.groups) + self.kernel_size
        weight_num_param = 1
        for wd in self.weight_shape:
            weight_num_param *= wd
        self.weight = torch.nn.Parameter(torch.Tensor(self.num_experts, weight_num_param))

        if bias:
            self.bias_shape = (self.out_channels,)
            self.bias = torch.nn.Parameter(torch.Tensor(self.num_experts, self.out_channels))
        else:
            self.register_parameter('bias', None)

        self.reset_parameters()

    def reset_parameters(self) -> None:
        init_weight = get_condconv_initializer(
            partial(nn.init.kaiming_uniform_, a=math.sqrt(5)), self.num_experts, self.weight_shape)
        init_weight(self.weight)
        if self.bias is not None:
            fan_in = np.prod(self.weight_shape[1:])
            bound = 1 / math.sqrt(fan_in)
            init_bias = get_condconv_initializer(
                partial(nn.init.uniform_, a=-bound, b=bound), self.num_experts, self.bias_shape)
            init_bias(self.bias)

    def forward(self, x: torch.Tensor, routing_weights: torch.Tensor) -> torch.Tensor:
        b, c, h, w = x.shape
        weight = torch.matmul(routing_weights, self.weight)
        new_weight_shape = (b * self.out_channels, self.in_channels // self.groups) + self.kernel_size
        weight = weight.view(new_weight_shape)
        bias = None
        if self.bias is not None:
            bias = torch.matmul(routing_weights, self.bias)
            bias = bias.view(b * self.out_channels)
        # move batch elements with channels so each batch element can be efficiently convolved with separate kernel
        x = x.view(1, b * c, h, w)
        if self.dynamic_padding:
            out = conv2d_same(
                x, weight, bias, stride=self.stride, padding=self.padding,
                dilation=self.dilation, groups=self.groups * b)
        else:
            out = F.conv2d(
                x, weight, bias, stride=self.stride, padding=self.padding,
                dilation=self.dilation, groups=self.groups * b)
        out = out.permute([1, 0, 2, 3]).view(b, self.out_channels, out.shape[-2], out.shape[-1])
        return out




# original code from https://github.com/krasserm/perceiver-io
# modified by Zilliz


from torch import nn

from towhee.models.perceiver.MultiHeadAttention import MultiHeadAttention


class CrossAttention(nn.Module):
    """
    Cross attention for Perceiver https://arxiv.org/pdf/2103.03206.pdf.
    Args:
        num_q_channels (`int`):
            Number of q channels.
        num_kv_channels (`int`):
            Number of k or v channels. k has the same channels as v.
        num_heads (`int`):
            Number of parallel attention heads.
        dropout (`nn.Module`):
            Dropout probability.
    """
    def __init__(self, num_q_channels: int, num_kv_channels: int, num_heads: int, dropout: float):
        super().__init__()
        self.q_norm = nn.LayerNorm(num_q_channels)
        self.kv_norm = nn.LayerNorm(num_kv_channels)
        self.attention = MultiHeadAttention(
            num_q_channels=num_q_channels, num_kv_channels=num_kv_channels, num_heads=num_heads, dropout=dropout
        )

    def forward(self, x_q, x_kv, pad_mask=None, attn_mask=None):
        x_q = self.q_norm(x_q)
        x_kv = self.kv_norm(x_kv)
        return self.attention(x_q, x_kv, pad_mask=pad_mask, attn_mask=attn_mask)



# Copyright 2021 Microsoft . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
import torch
from torch import nn
from typing import Optional
from towhee.models.utils.weight_init import trunc_normal_

class WindowAttention(nn.Module):
    r""" Window based multi-head self attention (W-MSA) module with relative position bias.
    It supports both of shifted and non-shifted window.
    Args:
        dim (int): Number of input channels.
        window_size (tuple[int]): The height and width of the window.
        num_heads (int): Number of attention heads.
        qkv_bias (bool, optional):  If True, add a learnable bias to query, key, value. Default: True
        attn_drop (float, optional): Dropout ratio of attention weight. Default: 0.0
        proj_drop (float, optional): Dropout ratio of output. Default: 0.0
    """

    def __init__(self, dim, window_size, num_heads, qkv_bias=True, attn_drop=0., proj_drop=0.):

        super().__init__()
        self.dim = dim
        self.window_size = window_size  # Wh, Ww
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim ** -0.5

        # define a parameter table of relative position bias
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros((2 * window_size[0] - 1) * (2 * window_size[1] - 1), num_heads))  # 2*Wh-1 * 2*Ww-1, nH

        # get pair-wise relative position index for each token inside the window
        coords_h = torch.arange(self.window_size[0])
        coords_w = torch.arange(self.window_size[1])
        coords = torch.stack(torch.meshgrid([coords_h, coords_w]))  # 2, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
        relative_coords[:, :, 0] += self.window_size[0] - 1  # shift to start from 0
        relative_coords[:, :, 1] += self.window_size[1] - 1
        relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1
        relative_position_index = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
        self.register_buffer('relative_position_index', relative_position_index)

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        trunc_normal_(self.relative_position_bias_table, std=.02)
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, x, mask: Optional[torch.Tensor] = None):
        """
        Args:
            x: input features with shape of (num_windows*B, N, C)
            mask: (0/-inf) mask with shape of (num_windows, Wh*Ww, Wh*Ww) or None
        """
        b_, n, c = x.shape
        qkv = self.qkv(x).reshape(b_, n, 3, self.num_heads, c // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]  # make torchscript happy (cannot use tensor as tuple)

        q = q * self.scale
        attn = (q @ k.transpose(-2, -1))

        relative_position_bias = self.relative_position_bias_table[self.relative_position_index.view(-1)].view(
            self.window_size[0] * self.window_size[1], self.window_size[0] * self.window_size[1], -1)  # Wh*Ww,Wh*Ww,nH
        relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()  # nH, Wh*Ww, Wh*Ww
        attn = attn + relative_position_bias.unsqueeze(0)

        if mask is not None:
            nw = mask.shape[0]
            attn = attn.view(b_ // nw, nw, self.num_heads, n, n) + mask.unsqueeze(1).unsqueeze(0)
            attn = attn.view(-1, self.num_heads, n, n)
            attn = self.softmax(attn)
        else:
            attn = self.softmax(attn)

        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2).reshape(b_, n, c)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x



# Copyright 2021  Facebook. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.


import torch
from torch import nn


class SpatialTemporalClsPositionalEncoding(nn.Module):
    """
    Add a cls token and apply a spatial-temporal encoding to a tensor.
    Args:
        embed_dim(int):
            Embedding dimension for input sequence.
        patch_embed_shape(Tuple):
            The number of patches in each dimension (T, H, W) after patch embedding.
        sep_pos_embed(bool):
            If set to true, one positional encoding is used for
            spatial patches and another positional encoding is used for temporal
            sequence. Otherwise, only one positional encoding is used for all the patches.
        has_cls(bool):
            If set to true, a cls token is added in the beginning of each input sequence.
    """

    def __init__(
        self,
        embed_dim,
        patch_embed_shape,
        sep_pos_embed=False,
        has_cls=True,
    ) -> None:
        super().__init__()
        self.embed_dim = embed_dim
        self.patch_embed_shape = patch_embed_shape
        self.sep_pos_embed = sep_pos_embed
        self.has_cls = has_cls
        assert (
            len(self.patch_embed_shape) == 3
        ), "Patch_embed_shape should be in the form of (T, H, W)."
        self.num_spatial_patch = patch_embed_shape[1] * patch_embed_shape[2]
        self.num_temporal_patch = patch_embed_shape[0]

        if self.has_cls:
            self.cls_token = nn.Parameter(torch.zeros(1, 1, self.embed_dim))
            num_patches = self.num_spatial_patch * self.num_temporal_patch + 1
        else:
            num_patches = self.num_spatial_patch * self.num_temporal_patch

        if self.sep_pos_embed:
            self.pos_embed_spatial = nn.Parameter(
                torch.zeros(1, self.num_spatial_patch, self.embed_dim)
            )
            self.pos_embed_temporal = nn.Parameter(
                torch.zeros(1, self.num_temporal_patch, self.embed_dim)
            )
            if self.has_cls:
                self.pos_embed_class = nn.Parameter(torch.zeros(1, 1, self.embed_dim))
        else:
            self.pos_embed = nn.Parameter(torch.zeros(1, num_patches, self.embed_dim))

    def get_patch_embed_shape(self):
        return self.patch_embed_shape

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x(torch.Tensor):
                Input tensor.
        """
        b, _, _ = x.shape
        if self.has_cls:
            cls_tokens = self.cls_token.expand(b, -1, -1)
            x = torch.cat((cls_tokens, x), dim=1)

        if self.sep_pos_embed:
            pos_embed = self.pos_embed_spatial.repeat(
                1, self.num_temporal_patch, 1
            ) + torch.repeat_interleave(
                self.pos_embed_temporal,
                self.num_spatial_patch,
                dim=1,
            )
            if self.has_cls:
                pos_embed = torch.cat([self.pos_embed_class, pos_embed], 1)
            x = x + pos_embed
        else:
            x = x + self.pos_embed

        return x



# Copyright 2021 Microsoft . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
import math
import torch
from torch import nn
from towhee.models.utils.weight_init import trunc_normal_
from towhee.models.layers.patch_embed2d import PatchEmbed2D
from towhee.models.layers.patch_merging import PatchMerging
from .basic_layer import BasicLayer
from .utils import init_vit_weights

class SwinTransformer(nn.Module):
    r""" Swin Transformer
        A PyTorch impl of : `Swin Transformer: Hierarchical Vision Transformer using Shifted Windows`  -
          https://arxiv.org/pdf/2103.14030
    Args:
        img_size (int | tuple(int)): Input image size. Default 224
        patch_size (int | tuple(int)): Patch size. Default: 4
        in_chans (int): Number of input image channels. Default: 3
        num_classes (int): Number of classes for classification head. Default: 1000
        embed_dim (int): Patch embedding dimension. Default: 96
        depths (tuple(int)): Depth of each Swin Transformer layer.
        num_heads (tuple(int)): Number of attention heads in different layers.
        window_size (int): Window size. Default: 7
        mlp_ratio (float): Ratio of mlp hidden dim to embedding dim. Default: 4
        qkv_bias (bool): If True, add a learnable bias to query, key, value. Default: True
        drop_rate (float): Dropout rate. Default: 0
        attn_drop_rate (float): Attention dropout rate. Default: 0
        drop_path_rate (float): Stochastic depth rate. Default: 0.1
        norm_layer (nn.Module): Normalization layer. Default: nn.LayerNorm.
        ape (bool): If True, add absolute position embedding to the patch embedding. Default: False
        patch_norm (bool): If True, add normalization after patch embedding. Default: True
        use_checkpoint (bool): Whether to use checkpointing to save memory. Default: False
    """

    def __init__(self, img_size=224, patch_size=4, in_chans=3, num_classes=1000,
                 embed_dim=96, depths=(2, 2, 6, 2), num_heads=(3, 6, 12, 24),
                 window_size=7, mlp_ratio=4., qkv_bias=True,
                 drop_rate=0., attn_drop_rate=0., drop_path_rate=0.1,
                 norm_layer=nn.LayerNorm, ape=False, patch_norm=True,
                 use_checkpoint=False, weight_init=''):
        super().__init__()

        self.num_classes = num_classes
        self.num_layers = len(depths)
        self.embed_dim = embed_dim
        self.ape = ape
        self.patch_norm = patch_norm
        self.num_features = int(embed_dim * 2 ** (self.num_layers - 1))
        self.mlp_ratio = mlp_ratio

        # split image into non-overlapping patches
        self.patch_embed = PatchEmbed2D(
            img_size=img_size, patch_size=patch_size, in_chans=in_chans, embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None)
        num_patches = self.patch_embed.num_patches
        self.patch_grid = self.patch_embed.grid_size

        # absolute position embedding
        if self.ape:
            self.absolute_pos_embed = nn.Parameter(torch.zeros(1, num_patches, embed_dim))
            trunc_normal_(self.absolute_pos_embed, std=.02)
        else:
            self.absolute_pos_embed = None

        self.pos_drop = nn.Dropout(p=drop_rate)

        # stochastic depth
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule

        # build layers
        layers = []
        for i_layer in range(self.num_layers):
            layers += [BasicLayer(
                dim=int(embed_dim * 2 ** i_layer),
                input_resolution=(self.patch_grid[0] // (2 ** i_layer), self.patch_grid[1] // (2 ** i_layer)),
                depth=depths[i_layer],
                num_heads=num_heads[i_layer],
                window_size=window_size,
                mlp_ratio=self.mlp_ratio,
                qkv_bias=qkv_bias,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr[sum(depths[:i_layer]):sum(depths[:i_layer + 1])],
                norm_layer=norm_layer,
                downsample=PatchMerging if (i_layer < self.num_layers - 1) else None,
                use_checkpoint=use_checkpoint)
            ]
        self.layers = nn.Sequential(*layers)

        self.norm = norm_layer(self.num_features)
        self.avgpool = nn.AdaptiveAvgPool1d(1)
        self.head = nn.Linear(self.num_features, num_classes) if num_classes > 0 else nn.Identity()

        assert weight_init in ('jax', 'jax_nlhb', 'nlhb', '')
        head_bias = -math.log(self.num_classes) if 'nlhb' in weight_init else 0.
        if weight_init.startswith('jax'):
            for n, m in self.named_modules():
                init_vit_weights(m, n, head_bias=head_bias, jax_impl=True)
        else:
            self.apply(init_vit_weights)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'absolute_pos_embed'}

    @torch.jit.ignore
    def no_weight_decay_keywords(self):
        return {'relative_position_bias_table'}

    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes):
        self.num_classes = num_classes
        self.head = nn.Linear(self.num_features, num_classes) if num_classes > 0 else nn.Identity()

    def forward_features(self, x):
        x = self.patch_embed(x)
        if self.absolute_pos_embed is not None:
            x = x + self.absolute_pos_embed
        x = self.pos_drop(x)
        x = self.layers(x)
        x = self.norm(x)  # B L C
        x = self.avgpool(x.transpose(1, 2))  # B C 1
        x = torch.flatten(x, 1)
        return x

    def forward(self, x):
        x = self.forward_features(x)
        x = self.head(x)
        return x



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Union


class Array:
    """
    One-dimensional array of data

    Args:
        data (`numpy.ndarray`, `list`, `dict` or scalar):
            Contains data to be stored in `Array`.
        dtype (`numpy.dtype`, `str`, `PIL.Image` or `towhee.DType`, optional):
            Data type of the `Array`. If not specified, it will be inferred from `data`.
        name (`str`):
            Name of the `Array`.
        copy (`bool`, default False):
            Copy the `data` contents to the `Array`.
    """

    def __init__(
        self,
        data=None,
        # dtype=None,
        name: str = None,
        # copy=False
    ):

        self._name = name

        # For `data` is `list`
        if isinstance(data, list):
            pass
        # For `data` is `None`
        elif data is None:
            data = []
        # For `data` is scalar
        else:
            data = [data]

        self._data = data
        self._offset = 0

    def __iter__(self):
        return self._data.__iter__()

    def __getitem__(self, key: Union[int, slice]):
        if isinstance(key, int):
            if key >= self._offset:
                return self._data[key - self._offset]
            elif key < 0:
                return self._data[len(self._data) + key]
            else:
                raise IndexError(f'element with index={key} has been released')
        elif isinstance(key, slice):
            if key.start >= self._offset:
                if key.stop is None:
                    stop = None
                else:
                    stop = key.stop - self._offset
                return self._data[key.start - self._offset:stop]
            else:
                raise IndexError(f'element with index={key.start} has been released' )
        else:
            raise IndexError('only integers are invalid indices')

    def __repr__(self):
        return self._data.__repr__()

    def __len__(self):
        """
        Number of elements in the `Array`.
        """
        return len(self._data)

    @property
    def size(self) -> int:
        """
        Total number of elements in the `Array` since creation.
        """
        return len(self._data) + self._offset

    @property
    def name(self) -> str:
        """
        Name of the Array'.
        """
        return self._name

    @property
    def data(self) -> List:
        """
        Data of the `Array`
        """
        return self._data

    def set_name(self, name):
        self._name = name

    def clear(self):
        self._data = []
        self._offset = 0

    def get_relative(self, key: int):
        return self._data[key + self._offset]

    def is_empty(self) -> bool:
        """
        Indicator whether `Array` is empty.

        Returns:
            (`bool`)
                Return `True` if `Array` has no elements.
        """
        return not self.size

    def put(self, item):
        """
        Append one item to the end of this `Array`.
        """
        self._data.append(item)

    def append(self, data):
        """
        Append a list-like items to the end of this `Array`.

        Args:
            data (`list` or `Array`):
                The data to be appended.
        """
        if isinstance(data, Array):
            self._data.extend(data.data)
        elif isinstance(data, List):
            self._data.extend(data)

   # TODO: Check out of bounds cases.
    def gc(self, offset):
        """
        Release the unreferenced lower part of the `Array` up to the index but not including.
        """

        if offset == -1:
            self.clear()
            return
        if offset > self.size:
            offset = self.size

        release_offset = offset - self._offset
        if release_offset > 0:
            del self._data[:release_offset]
            self._offset = offset

if __name__ == '__main__':
    arr = Array()
    arr.append([0, 1, 2, 3])
    arr.gc(4)
    arr.append([4])
    print(arr[4])
    print(arr)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# testable entrance
def main_body(args):
    import argparse  # pylint: disable=import-outside-toplevel

    parser = argparse.ArgumentParser('towhee', 'Towhee command line tool')
    subparsers = parser.add_subparsers(dest='action', help='towhee actions')

    from .subcmds.develop import DevelopCommand    # pylint: disable=import-outside-toplevel
    from .subcmds.execute import ExecuteCommand    # pylint: disable=import-outside-toplevel
    actions = {
        'develop': DevelopCommand,
        'execute': ExecuteCommand
    }

    for _, impl in actions.items():
        impl.install(subparsers)

    parsed = parser.parse_args(args)
    if parsed.action is None:
        parser.print_help()
    else:
        actions[parsed.action](parsed)()


# main entrance
def main():
    import sys  # pylint: disable=import-outside-toplevel
    main_body(sys.argv[1:])


if __name__ == '__main__':
    main()



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
from typing import Union

from torch import nn

from towhee.models.layers.conv2d_same import Conv2dSame
from towhee.models.layers.padding_functions import get_padding_value

def create_conv2d_pad(in_chs: int, out_chs: int, kernel_size: int, **kwargs) -> Union[nn.Conv2d, Conv2dSame]:
    padding = kwargs.pop('padding', '')
    kwargs.setdefault('bias', False)
    padding, is_dynamic = get_padding_value(padding, kernel_size, **kwargs)
    if is_dynamic:
        return Conv2dSame(in_chs, out_chs, kernel_size, **kwargs)
    else:
        if 'num_experts' in kwargs:
            kwargs.pop('num_experts', '')
        return nn.Conv2d(in_chs, out_chs, kernel_size, padding=padding, **kwargs)




# original code from https://github.com/krasserm/perceiver-io
# modified by Zilliz


import os

try:
    import fairscale
except ModuleNotFoundError:
    os.system("pip install fairscale")
from fairscale.nn import checkpoint_wrapper

from towhee.models.perceiver.Sequential import Sequential
from towhee.models.perceiver.Residual import Residual
from towhee.models.perceiver.SelfAttention import SelfAttention
from towhee.models.perceiver.mlp import mlp


def create_self_attention(num_channels: int, num_heads: int, dropout: float, activation_checkpoint: bool = False):
    """
    Self attention block for Perceiver https://arxiv.org/pdf/2103.03206.pdf.

    Args:
        num_channels (`int`):
            Number of q channels.
        num_heads (`int`):
            Number of parallel attention heads.
        dropout (`nn.Module`):
            Dropout probability.
        activation_checkpoint (`bool`):
            Use activation checkpointing.

    Return (`nn.Module`):
        Configured self attention layer.
    """
    layer = Sequential(
        Residual(SelfAttention(num_channels, num_heads, dropout), dropout), Residual(mlp(num_channels), dropout)
    )
    return layer if not activation_checkpoint else checkpoint_wrapper(layer)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import threading
from typing import Tuple

from towhee.engine.operator_loader import OperatorLoader
from towhee.hparam.hyperparameter import CallTracer, param_scope
# pylint: disable=unused-import
from towhee.hub import preclude


def op(operator_src: str, tag: str = 'main', **kwargs):
    """
    Entry method which takes either operator tasks or paths to python files or class in notebook.
    An `Operator` object is created with the init args(kwargs).
    Args:
        operator_src (`str`):
            operator name or python file location or class in notebook.
        tag (`str`):
            Which tag to use for operators on hub, defaults to `main`.
    Returns
        (`typing.Any`)
            The `Operator` output.
    """
    if isinstance(operator_src, type):
        class_op = type('operator', (operator_src, ), kwargs)
        return class_op.__new__(class_op, **kwargs)

    loader = OperatorLoader()
    if os.path.isfile(operator_src):
        op_obj = loader.load_operator_from_path(operator_src, kwargs)
    else:
        op_obj = loader.load_operator(operator_src, kwargs, tag)

    return op_obj


class _OperatorLazyWrapper:
    """
    operator wrapper for lazy initialization.
    """
    def __init__(self, real_name: str, index: Tuple[str], tag: str = 'main', **kws) -> None:
        self._name = real_name.replace('.', '/').replace('_', '-')
        self._index = index
        self._tag = tag
        self._kws = kws
        self._op = None
        self._lock = threading.Lock()

    def __call__(self, *arg, **kws):
        with self._lock:
            if self._op is None:
                self._create_op()

        if bool(self._index):
            # Multi inputs.
            if isinstance(self._index[0], tuple):
                args = []
                for i in self._index[0]:
                    args.append(getattr(arg[0], i))
                res = self._op(*args, **kws)
            # Single input.
            else:
                args = getattr(arg[0], self._index[0])
                res = self._op(args, **kws)

            # Multi outputs.
            if isinstance(res, tuple):
                if not isinstance(self._index[1], tuple) or len(self._index[1]) != len(res):
                    raise IndexError(f'Op has {len(res)} outputs, but {len(self._index[1])} indices are given.')
                for i in range(len(res)):
                    setattr(arg[0], self._index[1][i], res[i])
            # Single output.
            else:
                setattr(arg[0], self._index[1], res)

            return arg[0]
        else:
            res = self._op(*arg, **kws)
            return res

    def train(self, *arg, **kws):
        with self._lock:
            if self._op is None:
                self._create_op()
        return self._op.train(*arg, **kws)

    def fit(self, *arg):
        self._op.fit(*arg)

    @property
    def is_stateful(self):
        with self._lock:
            if self._op is None:
                self._create_op()
        return hasattr(self._op, 'fit')

    def set_state(self, state):
        with self._lock:
            if self._op is None:
                self._create_op()
        self._op.set_state(state)

    def set_training(self, flag):
        self._op.set_training(flag)

    @property
    def function(self):
        return self._name

    @property
    def init_args(self):
        return self._kws

    @staticmethod
    def callback(real_name: str, index: Tuple[str], *arg, **kws):
        if arg and not kws:
            raise ValueError('The init args should be passed in the form of kwargs(i.e. You should specify the keywords of your init arguments.)')
        if len(arg) == 0:
            return _OperatorLazyWrapper(real_name, index, **kws)
        else:
            return _OperatorLazyWrapper(real_name, index, arg[0], **kws)

    def _create_op(self):
        """
        Instantiate the operator.
        """
        # pylint: disable=unused-variable
        with param_scope(index=self._index) as hp:
            self._op = op(self._name, self._tag, **self._kws)


class OpsCallTracer(CallTracer):
    """
    Entry point for creating operator instances, for example:

    >>> op_instance = ops.my_namespace.my_operator_name(init_arg1=xxx, init_arg2=xxx)

    An instance of `my_namespace`/`my_operator_name` is created.
    """
    def __init__(self, callback=None, path=None, index=None):
        super().__init__(callback=callback, path=path, index=index)


ops = OpsCallTracer(_OperatorLazyWrapper.callback)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.models.layers.activations.gelu import GELU, gelu
from towhee.models.layers.activations.hardsigmoid import HardSigmoid, hard_sigmoid
from towhee.models.layers.activations.hardswish import HardSwish, hard_swish
from towhee.models.layers.activations.mish import Mish, mish
from towhee.models.layers.activations.hardmish import HardMish, hard_mish
from towhee.models.layers.activations.prelu import PReLU
from towhee.models.layers.activations.sigmoid import Sigmoid, sigmoid
from towhee.models.layers.activations.swish import Swish, swish
from towhee.models.layers.activations.tanh import Tanh, tanh




# original code from https://github.com/krasserm/perceiver-io
# modified by Zilliz
from torch.nn import Sequential
from torch import nn


def mlp(num_channels: int):
    return Sequential(
        nn.LayerNorm(num_channels),
        nn.Linear(num_channels, num_channels),
        nn.GELU(),
        nn.Linear(num_channels, num_channels),
    )



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import PosixPath
from typing import Union
import numpy as np

from towhee.utils.log import engine_log
from towhee.types import Image

try:
    from PIL import Image as PILImage
except ModuleNotFoundError as e:
    engine_log.error('PIL not found, you can install via `pip install pillow`.')
    raise ModuleNotFoundError('PIL not found, you can install via `pip install pillow`.') from e


def from_src(src: Union[str, PosixPath]) -> Image:
    """
    Load the image from url/path as towhee's Image object.

    Args:
        src (`Union[str, path]`):
            The path leads to the image.

    Returns:
        (`towhee.types.Image`)
            The image wrapepd as towhee's Image.
    """
    pil_img = PILImage.open(src)
    img = from_pil(pil_img)

    return img


def from_pil(pil_img: PILImage.Image) -> Image:
    '''
    Convert a PIL.Image.Image into towhee.types.Image.

    Args:
        pil_img (`PIL.Image.Image`):
            A PIL image.

    Returns:
        (`towhee.types.Image`)
            The image wrapepd as towhee Image.
    '''
    img_bytes = pil_img.tobytes()
    img_width = pil_img.width
    img_height = pil_img.height
    img_channel = len(pil_img.split())
    img_mode = pil_img.mode
    img_array = np.array(pil_img)

    towhee_img = Image(img_bytes, img_width, img_height, img_channel, img_mode, img_array)

    return towhee_img


def to_pil(towhee_img: Image) -> PILImage.Image:
    """
    Convert a towhee.types.Image into PIL.Image.Image.

    Args:
        towhee_img (`towhee.types.Image`):
            A towhee image.

    Returns:
        (`PIL.Image.Image`)
            A PIL image.
    """
    mode = towhee_img.mode
    size = (towhee_img.width, towhee_img.height)
    data = towhee_img.image

    pil_img = PILImage.frombytes(mode, size, data)

    return pil_img



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
from typing import Dict, Tuple
from abc import ABC
from enum import Enum, auto
from threading import Event
import traceback

from towhee.types._frame import FRAME, _Frame
from towhee.engine.status import Status
from towhee.utils.log import engine_log
from towhee.hparam import param_scope


class RunnerStatus(Enum):
    IDLE = auto()
    RUNNING = auto()
    FINISHED = auto()
    FAILED = auto()

    @staticmethod
    def is_end(status: 'RunnerStatus') -> bool:
        return status in [RunnerStatus.FINISHED, RunnerStatus.FAILED]


class _OpInfo:
    """
    Operator hub info.
    """

    def __init__(self, op_name: str, hub_op_id: str, op_args: Dict[str, any], tag: str) -> None:
        self._op_name = op_name
        self._hub_op_id = hub_op_id
        self._tag = tag
        self._op_args = op_args

    @property
    def op_name(self):
        return self._op_name

    @property
    def hub_op_id(self):
        return self._hub_op_id

    @property
    def op_args(self):
        return self._op_args

    @property
    def tag(self):
        return self._tag


class RunnerBase(ABC):
    """
    Running instance of op.

    The base class provides some function to control status.
    """

    def __init__(
        self,
        name: str,
        index: int,
        op_name: str,
        tag: str,
        hub_op_id: str,
        op_args: Dict[str, any],
        readers=None,
        writer=None,
    ) -> None:
        self._name = name
        self._index = index
        self._status = RunnerStatus.IDLE
        self._msg = None
        self._op_info = _OpInfo(op_name, hub_op_id, op_args, tag)

        self._readers = readers
        self._sleep_time = 0

        # only concat can have multiple readers
        if self._readers:
            self._reader = readers[0]
        self._writer = writer
        self._need_stop = False
        self._end_event = Event()
        self._op = None

    def _set_end_status(self, status: RunnerStatus):
        self._set_status(status)
        self._end_event.set()
        engine_log.info('%s ends with status: %s', self._name, status)

    @property
    def op_name(self):
        return self._op_info.op_name

    @property
    def hub_op_id(self):
        return self._op_info.hub_op_id

    @property
    def tag(self):
        return self._op_info.tag

    @property
    def op_args(self):
        return self._op_info.op_args

    @property
    def status(self) -> RunnerStatus:
        return self._status

    @property
    def msg(self):
        return self._msg

    @property
    def op(self):
        return self._op

    def is_end(self) -> bool:
        return RunnerStatus.is_end(self.status)

    def set_op(self, op):
        self._op = op

    def unset_op(self):
        self._op = None

    def _set_finished(self) -> None:
        self._set_end_status(RunnerStatus.FINISHED)

    def _set_idle(self) -> None:
        self._set_status(RunnerStatus.IDLE)

    def _set_failed(self, msg: str) -> None:
        error_info = '{} runs failed, error msg: {}'.format(str(self), msg)
        self._msg = error_info
        self._set_end_status(RunnerStatus.FAILED)

    def _set_running(self) -> None:
        self._set_status(RunnerStatus.RUNNING)

    def _set_status(self, status: RunnerStatus) -> None:
        self._status = status

    def is_idle(self) -> bool:
        return self.status == RunnerStatus.IDLE

    def is_finished(self) -> bool:
        return self.status == RunnerStatus.FINISHED

    def set_stop(self) -> None:
        engine_log.info('Begin to stop %s', str(self))
        self._need_stop = True
        for reader in self._readers:
            reader.close()

    def __str__(self) -> str:
        return '{}:{}'.format(self._name, self._index)

    def _call_op(self, inputs) -> Status:
        try:
            with param_scope() as hp:
                hp().towhee.need_schema = True
                outputs = self._op(**inputs)
            return Status.ok_status(outputs)
        except Exception as e:  # pylint: disable=broad-except
            err = '{}, {}'.format(str(e), traceback.format_exc())
            return Status.err_status(err)

    def _get_inputs(self) -> Tuple[bool, Dict[str, any]]:
        try:
            data, self._row_data = self._reader.read()
            assert isinstance(self._row_data[-1], _Frame)
            self._frame = self._row_data[-1]
            self._frame.prev_id = self._frame.row_id
            return False, data
        except StopIteration:
            return True, None

    def _set_outputs(self, output: any):
        if hasattr(output, '_asdict'):
            output = output._asdict()

        if isinstance(output, dict):
            output.update({FRAME: self._frame})

        self._writer.write(output)

    def process_step(self) -> bool:
        is_end, op_input_params = self._get_inputs()
        if is_end:
            self._set_finished()
            return True

        if op_input_params is None:
            # No data in dataframe, but dataframe is not sealed
            self._set_idle()
            return True

        st = self._call_op(op_input_params)
        if st.is_ok():
            self._set_outputs(st.data)
            return False
        else:
            self._set_failed(st.msg)
            return True

    def slow_down(self, time_sec: float):
        self._sleep_time = time_sec

    def speed_up(self):
        self._sleep_time = 0

    def sleep(self):
        if self._sleep_time > 0:
            time.sleep(self._sleep_time)

    def join(self):
        self._end_event.wait()

    def process(self):
        engine_log.info('Begin to run %s', str(self))
        self._set_running()
        while True:
            if not self._need_stop:
                try:
                    if self.process_step():
                        break
                except Exception as e:  # pylint: disable=broad-except
                    err = '{}, {}'.format(e, traceback.format_exc())
                    self._set_failed(err)
                    break
            else:
                self._set_finished()
                break
            self.sleep()



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.hparam import param_scope
from towhee.hparam import HyperParameter


class StateMixin:
    """
    Mixin for state tracking.

    Examples:

    >>> from towhee import DataCollection, State
    >>> from towhee import param_scope
    >>> dc = DataCollection.range(10).set_state(State(a=1))
    >>> dc.get_state()
    {'a': 1}

    >>> dc = dc.map(lambda x: x+1).map(lambda x: x*2)
    >>> dc.get_state()
    {'a': 1}
    """

    def __init__(self):
        super().__init__()
        with param_scope() as hp:
            parent = hp().data_collection.parent(None)
        if parent is not None:
            self.set_state(parent.get_state())

    def get_state(self):
        """
        Get the state storage for `DataCollection`

        Returns:
            State: the state storage
        """
        if hasattr(self, '_state') and isinstance(self._state, HyperParameter):
            return self._state
        return None

    def set_state(self, state):
        """
        Set the state storage for `DataCollection`

        Args:
            state (State): state storage

        Returns:
            DataCollection: data collection itself
        """
        self._state = state
        return self

    def set_training(self, state=None):
        """
        Set training mode for stateful operators

        Args:
            state (State, optional): Update the state storage. Defaults to None.

        Returns:
            DataCollection: data collection itself
        """
        if state is not None:
            self._state = state
        if self.get_state() is None:
            self._state = HyperParameter()
        self._state.__mode__ = 'training'
        return self

    def set_evaluating(self, state=None):
        """
        Set evaluating mode for stateful operators

        Args:
            state (State, optional): Update the state storage. Defaults to None.

        Returns:
            DataCollection: data collection itself
        """
        if state is not None:
            self._state = state
        if self.get_state() is None:
            self._state = HyperParameter()
        self._state.__mode__ = 'evaluating'
        return self

    def smap(self, op):
        op.set_state(self._state)
        if self._state().__mode__('evaluating') == 'training':
            op.set_training(True)
            with param_scope() as hp:
                hp().towhee.data_collection.training = True
                for x in self._iterable:
                    op(x)
            op.set_training(False)
            op.fit()
        return self._factory(map(op, self._iterable))


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from towhee.engine import register
from towhee.utils.log import engine_log
from towhee.operator.stateful_operator import StatefulOperator
try:
    from scipy import sparse
except ModuleNotFoundError as e:
    engine_log.error('scipy not found, you can install via `pip install scipy`.')
    raise ModuleNotFoundError('scipy not found, you can install via `pip install scipy`.') from e


# pylint: disable=import-outside-toplevel
# pylint: disable=invalid-name
@register(name='builtin/logistic_regression')
class logistic_regression(StatefulOperator):
    """
    Logistic Regression model encapsulate as an Operator.
    """
    def __init__(self, name: str = None, **kws):
        super().__init__(name=name)
        self._model_agrs = kws

    def fit(self):
        from towhee.utils.sklearn_utils import LogisticRegression
        X = sparse.vstack(self._data[0])
        y = np.array(self._data[1]).reshape([-1, 1])
        self._state.model = LogisticRegression(**self._model_agrs)
        self._state.model.fit(X, y)

    def predict(self, *arg):
        return self._state.model.predict(arg[0])


@register(name='builtin/random_forest')
class random_forest(StatefulOperator):
    """
    Random Forest Classifier model encapsulate as an Operator.
    """
    def __init__(self, name: str = None, **kws):
        super().__init__(name=name)
        self._model_agrs = kws

    def fit(self):
        from towhee.utils.sklearn_utils import RandomForestClassifier
        X = sparse.vstack(self._data[0])
        y = np.array(self._data[1]).reshape([-1, 1])
        self._state.model = RandomForestClassifier(**self._model_agrs)
        self._state.model.fit(X, y)

    def predict(self, *arg):
        return self._state.model.predict(arg[0])


@register(name='builtin/decision_tree')
class decision_tree(StatefulOperator):
    """
    Decision Tree Classifier model encapsulate as an Operator.
    """
    def __init__(self, name: str = None, **kws):
        super().__init__(name=name)
        self._model_agrs = kws

    def fit(self):
        from towhee.utils.sklearn_utils import DecisionTreeClassifier
        X = sparse.vstack(self._data[0])
        y = np.array(self._data[1]).reshape([-1, 1])
        self._state.model = DecisionTreeClassifier(**self._model_agrs)
        self._state.model.fit(X, y)

    def predict(self, *arg):
        return self._state.model.predict(arg[0])


@register(name='builtin/svc')
class svc(StatefulOperator):
    """
    SVM Classifier model encapsulate as an Operator.
    """
    def __init__(self, name: str = None, **kws):
        super().__init__(name=name)
        self._model_agrs = kws

    def fit(self):
        from towhee.utils.sklearn_utils import svm
        X = sparse.vstack(self._data[0])
        y = np.array(self._data[1]).reshape([-1, 1])
        self._state.model = svm.SVC(**self._model_agrs)
        self._state.model.fit(X, y)

    def predict(self, *arg):
        return self._state.model.predict(arg[0])



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import wraps
from typing import Callable


def arg(arg_idx: int, preprocess: Callable):
    """
    Type utils apply to function arguments.
    Args:
        arg_idx (`int`):
            The index number of the applied argument.
        preprocess (`Callable`):
            The arg preprocess function.
    """
    def decorate(func):
        @ wraps(func)
        def wrapper(*args, **kwargs):
            x = preprocess(args[arg_idx])
            new_args = args[:arg_idx] + (x,) + args[arg_idx + 1:]
            return func(*new_args, **kwargs)
        return wrapper
    return decorate


# pylint: disable=invalid-name
class to_image_color:
    """
    convert images from one color-space to another, like BGR ↔ Gray, BGR ↔ HSV, etc.
    Args:
        mode (`str`):
            The target color space type, for example, 'RGB'.

    Returns (`numpy.ndarray`):
        The coverted image.

    Raise:
        ValueError:
            ValueError occurs if the color conversion is not supported.

    Examples:

    >>> import numpy as np
    >>> from towhee.operator import PyOperator
    >>> from towhee.types import Image, arg, to_image_color
    >>> class ExtractRed(PyOperator):
    ...     def __init__(self) -> None:
    ...         super().__init__()
    ...     @arg(0, to_image_color('BGR'))
    ...     def __call__(self, img):
    ...         red_img = Image(np.zeros(img.shape), 'BGR')
    ...         red_img[:,:,2] = img[:,:,2]
    ...         return red_img
    """

    def __init__(self, mode: str):
        self._mode = mode

    def __call__(self, img):
        # pylint: disable=import-outside-toplevel
        from towhee.types import image_utils

        return image_utils.to_image_color(img, self._mode)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import sys
from pathlib import Path
from typing import Any, Dict, Union

from towhee.operator import Operator
from towhee.operator.nop import NOPOperator
from towhee.operator.concat_operator import ConcatOperator
from towhee.engine import LOCAL_OPERATOR_CACHE

from .operator_registry import OperatorRegistry
from towhee.hub.file_manager import FileManager

from towhee.hparam import param_scope


class OperatorLoader:
    """Wrapper class used to load operators from either local cache or a remote
    location.

    Args:
        cache_path: (`str`)
            Local cache path to use. If not specified, it will default to
            `$HOME/.towhee/operators`.
    """
    def __init__(self, cache_path: str = None):
        if cache_path is None:
            self._cache_path = LOCAL_OPERATOR_CACHE
        else:
            self._cache_path = Path(cache_path)

    def _load_interal_op(self, op_name: str, args: Dict[str, any]):
        if op_name in ['_start_op', '_end_op']:
            return NOPOperator()
        elif op_name == '_concat':
            return ConcatOperator(**args)
        else:
            # Not a interal operator
            return None

    def load_operator_from_internal(self, function: str, args: Dict[str, Any], tag: str) -> Operator:  # pylint: disable=unused-argument
        return self._load_interal_op(function, args)

    def load_operator_from_registry(self, function: str, args: Dict[str, Any], tag: str) -> Operator:  # pylint: disable=unused-argument
        op = OperatorRegistry.resolve(function)
        return self.instance_operator(op, args) if op is not None else None

    def load_operator_from_packages(self, function: str, args: Dict[str, Any], tag: str) -> Operator:  # pylint: disable=unused-argument
        try:
            module, fname = function.split('/')
            fname = fname.replace('-', '_')
            op_cls = ''.join(x.capitalize() or '_' for x in fname.split('_'))

            # module = '.'.join([module, fname, fname])
            module = '.'.join(['towheeoperator', '{}_{}'.format(module, fname), fname])
            op = getattr(importlib.import_module(module), op_cls)
            return self.instance_operator(op, args) if op is not None else None
        except Exception:  # pylint: disable=broad-except
            with param_scope() as hp:
                if hp().towhee.hub.use_pip(False):
                    return None  # TODO: download and install pip package from hub
                else:
                    return None

    def load_operator_from_path(self, path: Union[str, Path], args: Dict[str, Any]) -> Operator:
        """
        Load operator form local path.
        Args:
            path (`Union[str, Path]`):
                Path to the operator python file.
            args (`Dict[str, Any]`):
                The init args for OperatorClass.
        Returns
            (`typing.Any`)
                The `Operator` output.
        """
        path = Path(path)
        fname = Path(path).stem
        modname = 'towhee.operator.' + fname
        try:
            # support for ver1 operator API
            spec = importlib.util.spec_from_file_location(modname, path.resolve())

            # Create the module and then execute the module in its own namespace.
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)

            # Instantiate the operator object and return it to the caller for
            # `load_operator`. By convention, the operator class is simply the CamelCase
            # version of the snake_case operator.
            op_cls = ''.join(x.capitalize() or '_' for x in fname.split('_'))
            op = getattr(module, op_cls)
        except Exception:  # pylint: disable=broad-except
            # support for ver2 operator API
            path = path.parent / '__init__.py'
            spec = importlib.util.spec_from_file_location(modname, path.resolve())
            module = importlib.util.module_from_spec(spec)
            sys.modules[modname] = module
            spec.loader.exec_module(module)
            op = getattr(module, fname)
        return self.instance_operator(op, args) if op is not None else None

    def load_operator_from_cache(self, function: str, args: Dict[str, Any], tag: str) -> Operator:  # pylint: disable=unused-argument
        if '/' not in function:
            function = 'towhee/'+function
        try:
            fm = FileManager()
            path = fm.get_operator(operator=function, tag=tag)
        except ValueError as e:
            raise ValueError('operator {} not found!'.format(function)) from e

        if path is None:
            raise FileExistsError('Cannot find operator.')

        return self.load_operator_from_path(path, args)

    def load_operator(self, function: str, args: Dict[str, Any], tag: str) -> Operator:
        """Attempts to load an operator from cache. If it does not exist, looks up the
        operator in a remote location and downloads it to cache instead. By standard
        convention, the operator must be called `Operator` and all associated data must
        be contained within a single directory.

        Args:
            function: (`str`)
                Origin and method/class name of the operator. Used to look up the proper
                operator in cache.
        Raises:
            FileExistsError
                Cannot find operator.
        """

        for factory in [self.load_operator_from_internal,
                        self.load_operator_from_registry,
                        self.load_operator_from_packages,
                        self.load_operator_from_cache]:
            op = factory(function, args, tag)
            if op is not None:
                return op
        return None

    def instance_operator(self, op, args: Dict[str, Any]) -> Operator:
        with param_scope() as hp:
            if hp().towhee.dry_run(False):
                return NOPOperator()
            else:
                return op(**args) if args is not None else op()



# Copyright 2021 Microsoft . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
import torch
from torch import nn

class PatchMerging(nn.Module):
    r""" Patch Merging Layer.
    Args:
        input_resolution (tuple[int]): Resolution of input feature.
        dim (int): Number of input channels.
        norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
    """

    def __init__(self, input_resolution, dim, norm_layer=nn.LayerNorm):
        super().__init__()
        self.input_resolution = input_resolution
        self.dim = dim
        self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
        self.norm = norm_layer(4 * dim)

    def forward(self, x):
        """
        x: B, H*W, C
        """
        h, w = self.input_resolution
        b, l, c = x.shape
        assert l == h * w, 'input feature has wrong size'
        assert h % 2 == 0 and w % 2 == 0, f'x size ({h}*{w}) are not even.'

        x = x.view(b, h, w, c)

        x0 = x[:, 0::2, 0::2, :]  # B H/2 W/2 C
        x1 = x[:, 1::2, 0::2, :]  # B H/2 W/2 C
        x2 = x[:, 0::2, 1::2, :]  # B H/2 W/2 C
        x3 = x[:, 1::2, 1::2, :]  # B H/2 W/2 C
        x = torch.cat([x0, x1, x2, x3], -1)  # B H/2 W/2 4*C
        x = x.view(b, -1, 4 * c)  # B H/2*W/2 4*C

        x = self.norm(x)
        x = self.reduction(x)

        return x

    def extra_repr(self) -> str:
        return f'input_resolution={self.input_resolution}, dim={self.dim}'

    def flops(self):
        h, w = self.input_resolution
        flops = h * w * self.dim
        flops += (h // 2) * (w // 2) * 4 * self.dim * 2 * self.dim
        return flops



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee._types import Image


def to_image_color(img: Image, target_mode: str):
    """
    convert images from one color-space to another, like BGR ↔ Gray, BGR ↔ HSV, etc.
    """
    # pylint: disable=import-outside-toplevel
    from towhee.utils.cv2_utils import cv2

    if img.mode == target_mode:
        return img

    flag_name = 'COLOR_' + img.mode.upper() + '2' + target_mode.upper()
    flag = getattr(cv2, flag_name, None)

    if flag is None:
        raise ValueError('Can not convert image from %s to %s.' % (img.mode, target_mode))

    return Image(cv2.cvtColor(img, flag), target_mode.upper())


def from_pil(pil_img):
    '''
    Convert a PIL.Image.Image into towhee.types.Image.

    Args:
        pil_img (`PIL.Image.Image`):
            A PIL image.

    Returns:
        (`towhee.types.Image`)
            The image wrapepd as towhee Image.
    '''
    # pylint: disable=import-outside-toplevel
    import numpy as np

    return Image(np.array(pil_img), pil_img.mode)


def to_pil(img: Image):
    """
    Convert a towhee.types.Image into PIL.Image.Image.

    Args:
        img (`towhee.types.Image`):
            A towhee image.

    Returns (`PIL.Image.Image`):
            A PIL image.
    """
    # pylint: disable=import-outside-toplevel
    from PIL import Image as PILImage

    return PILImage.fromarray(img, img.mode)



# original code from https://github.com/SwinTransformer/Video-Swin-Transformer
# modified by Zilliz.

from torch import nn
import torch
from einops import rearrange


from towhee.models.layers.patch_embed3d import PatchEmbed3D
from towhee.models.layers.patch_merging3d import PatchMerging3D
from towhee.models.video_swin_transformer.video_swin_transformer_block import VideoSwinTransformerBlock


class VideoSwinTransformer(nn.Module):
    """
    Video Swin Transformer.
    Ze Liu, Jia Ning, Yue Cao, Yixuan Wei, Zheng Zhang, Stephen Lin, Han Hu
    https://arxiv.org/pdf/2106.13230.pdf

    Args:
        pretrained (`str`):
            Load pretrained weights. Default: None
        pretrained2d (`bool`):
            Load image pretrained weights. Default: True
        patch_size (`tuple[int]`):
            Patch size. Default: (4,4,4).
        in_chans (`int)`:
            Number of input image channels. Default: 3.
        embed_dim (`int`):
            Number of linear projection output channels. Default: 96.
        depths (`tuple[int]`):
            Depths of each Swin Transformer stage.
        num_heads (`tuple[int]`):
            Number of attention head of each stage.
        window_size (`int`):
            Window size. Default: 7.
        mlp_ratio (`float`):
            Ratio of mlp hidden dim to embedding dim. Default: 4.
        qkv_bias (`bool`):
            If True, add a learnable bias to query, key, value. Default: True
        qk_scale (`float`):
            Override default qk scale of head_dim ** -0.5 if set.
        drop_rate (`float`):
            Dropout rate.
        attn_drop_rate (`float`):
            Attention dropout rate. Default: 0.
        drop_path_rate (`float`):
            Stochastic depth rate. Default: 0.2.
        norm_layer (`nn.Module`):
            Normalization layer. Default: nn.LayerNorm.
        patch_norm (`bool`):
            If True, add normalization after patch embedding. Default: False.
        frozen_stages (`int`):
            Stages to be frozen (stop grad and set eval mode).
            -1 means not freezing any parameters.
        use_checkpoint (`bool`):
            Use checkpoint.
    """

    def __init__(self,
                 pretrained=None,
                 pretrained2d=True,
                 patch_size=(4, 4, 4),
                 in_chans=3,
                 embed_dim=96,
                 depths=(2, 2, 6, 2),
                 num_heads=(3, 6, 12, 24),
                 window_size=(2, 7, 7),
                 mlp_ratio=4.,
                 qkv_bias=True,
                 qk_scale=None,
                 drop_rate=0.,
                 attn_drop_rate=0.,
                 drop_path_rate=0.2,
                 norm_layer=nn.LayerNorm,
                 patch_norm=False,
                 frozen_stages=-1,
                 use_checkpoint=False):
        super().__init__()

        self.pretrained = pretrained
        self.pretrained2d = pretrained2d
        self.num_layers = len(depths)
        self.embed_dim = embed_dim
        self.patch_norm = patch_norm
        self.frozen_stages = frozen_stages
        self.window_size = window_size
        self.patch_size = patch_size

        # split image into non-overlapping patches
        self.patch_embed = PatchEmbed3D(
            patch_size=patch_size, c=in_chans, embed_dim=embed_dim,
            norm_layer=norm_layer if self.patch_norm else None)

        self.pos_drop = nn.Dropout(p=drop_rate)

        # stochastic depth
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule

        # build layers
        self.layers = nn.ModuleList()
        for i_layer in range(self.num_layers):
            layer = VideoSwinTransformerBlock(
                dim=int(embed_dim * 2**i_layer),
                depth=depths[i_layer],
                num_heads=num_heads[i_layer],
                window_size=window_size,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                qk_scale=qk_scale,
                drop=drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr[sum(depths[:i_layer]):sum(depths[:i_layer + 1])],
                norm_layer=norm_layer,
                downsample=PatchMerging3D if i_layer < self.num_layers-1 else None,
                use_checkpoint=use_checkpoint)
            self.layers.append(layer)

        self.num_features = int(embed_dim * 2**(self.num_layers-1))

        # add a norm layer for each output
        self.norm = norm_layer(self.num_features)

        self._freeze_stages()

    def _freeze_stages(self):
        if self.frozen_stages >= 0:
            self.patch_embed.eval()
            for param in self.patch_embed.parameters():
                param.requires_grad = False

        if self.frozen_stages >= 1:
            self.pos_drop.eval()
            for i in range(0, self.frozen_stages):
                m = self.layers[i]
                m.eval()
                for param in m.parameters():
                    param.requires_grad = False

    def inflate_weights(self, logger):
        """
        Inflate the swin2d parameters to swin3d.
        The differences between swin3d and swin2d mainly lie in an extra
        axis. To utilize the pretrained parameters in 2d model,
        the weight of swin2d models should be inflated to fit in the shapes of
        the 3d counterpart.
        Args:
            logger (`logging.Logger`):
                The logger used to print debugging infomation.
        """
        checkpoint = torch.load(self.pretrained, map_location="cpu")
        state_dict = checkpoint["model"]

        # delete relative_position_index since we always re-init it
        relative_position_index_keys = [k for k in state_dict.keys() if "relative_position_index" in k]
        for k in relative_position_index_keys:
            del state_dict[k]

        # delete attn_mask since we always re-init it
        attn_mask_keys = [k for k in state_dict.keys() if "attn_mask" in k]
        for k in attn_mask_keys:
            del state_dict[k]

        state_dict["patch_embed.proj.weight"] =\
            state_dict["patch_embed.proj.weight"].unsqueeze(2).repeat(1, 1, self.patch_size[0], 1, 1) \
            / self.patch_size[0]

        # bicubic interpolate relative_position_bias_table if not match
        relative_position_bias_table_keys = [k for k in state_dict.keys() if "relative_position_bias_table" in k]
        for k in relative_position_bias_table_keys:
            relative_position_bias_table_pretrained = state_dict[k]
            relative_position_bias_table_current = self.state_dict()[k]
            l1, nh1 = relative_position_bias_table_pretrained.size()
            l2, nh2 = relative_position_bias_table_current.size()
            l2 = (2*self.window_size[1]-1) * (2*self.window_size[2]-1)
            wd = self.window_size[0]
            if nh1 != nh2:
                logger.warning(f"Error in loading {k}, passing")
            else:
                if l1 != l2:
                    s1 = int(l1 ** 0.5)
                    relative_position_bias_table_pretrained_resized = torch.nn.functional.interpolate(
                        relative_position_bias_table_pretrained.permute(1, 0).view(1, nh1, s1, s1),
                        size=(2*self.window_size[1]-1, 2*self.window_size[2]-1),
                        mode="bicubic")
                    relative_position_bias_table_pretrained =\
                        relative_position_bias_table_pretrained_resized.view(nh2, l2).permute(1, 0)
            state_dict[k] = relative_position_bias_table_pretrained.repeat(2*wd-1, 1)

        msg = self.load_state_dict(state_dict, strict=False)
        logger.info(msg)
        logger.info(f"=> loaded successfully '{self.pretrained}'")
        del checkpoint
        torch.cuda.empty_cache()

    def forward(self, x):
        x = self.patch_embed(x)

        x = self.pos_drop(x)

        for layer in self.layers:
            x = layer(x.contiguous())

        x = rearrange(x, "n c d h w -> n d h w c")
        x = self.norm(x)
        x = rearrange(x, "n d h w c -> n c d h w")

        return x



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.engine import register
from towhee.operator.stateful_operator import StatefulOperator
# pylint: disable=import-outside-toplevel
# pylint: disable=useless-super-delegation
# pylint: disable=invalid-name


@register(name='builtin/standard_scaler')
class standard_scaler(StatefulOperator):
    """
    Standardize numerical features by removing the mean and scaling to unit variance.

    Examples:

    >>> from towhee import DataCollection, Entity
    >>> dc = (
    ...     DataCollection.range(10).map(lambda x: Entity(a=x))
    ...         .set_training()
    ...         .standard_scaler['a', 'b'](name='standard_scaler')
    ... )

    >>> [int(x.b*10) for x in dc.to_list()]
    [-15, -12, -8, -5, -1, 1, 5, 8, 12, 15]
    """
    def __init__(self, name: str = None):
        super().__init__(name)

    def fit(self):
        from sklearn.preprocessing import StandardScaler
        import numpy as np
        data = np.array(self._data[0]).reshape([-1, 1])
        self._state.model = StandardScaler()
        self._state.model.fit(data)

    def predict(self, x):
        import numpy as np
        data = np.array([x]).reshape([-1, 1])
        return self._state.model.transform(data)


@register(name='builtin/num_discretizer')
class num_discretizer(StatefulOperator):
    """
    Bin numerical features into intervals.

    Examples:

    >>> from towhee import DataCollection, Entity
    >>> dc = (
    ...     DataCollection.range(10).map(lambda x: Entity(a=x))
    ...         .set_training()
    ...         .num_discretizer['a', 'b'](name='discretizer', n_bins=3)
    ... )

    >>> [x.b.nonzero()[1][0] for x in dc.to_list()]
    [0, 0, 0, 1, 1, 1, 2, 2, 2, 2]
    """
    def __init__(self, name: str = None, n_bins=10, encode='onehot', strategy='quantile'):
        super().__init__(name)
        self._n_bins = n_bins
        self._encode = encode
        self._strategy = strategy

    def fit(self):
        from sklearn.preprocessing import KBinsDiscretizer
        import numpy as np
        data = np.array(self._data[0]).reshape([-1, 1])
        self._state.model = KBinsDiscretizer(n_bins=self._n_bins,
                                             encode=self._encode,
                                             strategy=self._strategy)
        self._state.model.fit(data)

    def predict(self, x):
        import numpy as np
        data = np.array([x]).reshape([-1, 1])
        return self._state.model.transform(data)


@register(name='builtin/cate_one_hot_encoder')
class cate_one_hot_encoder(StatefulOperator):
    """
    Standardize numerical features by removing the mean and scaling to unit variance.

    Examples:

    >>> from towhee import DataCollection, Entity
    >>> dc = (
    ...     DataCollection(['a','b','c','a','b']).map(lambda x: Entity(a=x))
    ...         .set_training()
    ...         .cate_one_hot_encoder['a', 'b'](name='one_hot_encoder')
    ... )

    >>> [x.b.nonzero()[1][0] for x in dc.to_list()]
    [0, 1, 2, 0, 1]
    """
    def __init__(self, name: str = None):
        super().__init__(name)

    def fit(self):
        from sklearn.preprocessing import OneHotEncoder
        import numpy as np
        data = np.array(self._data[0]).reshape([-1, 1])
        self._state.model = OneHotEncoder()
        self._state.model.fit(data)

    def predict(self, x):
        import numpy as np
        data = np.array([x]).reshape([-1, 1])
        return self._state.model.transform(data)


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False)



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import Any, Dict, NamedTuple

from towhee.operator import PyOperator


class NOPOperator(PyOperator):
    """No-op operator. Input arguments are redefined as a `NamedTuple` and returned as
    outputs.
    """

    def __init__(self):
        #pylint: disable=useless-super-delegation
        super().__init__()

    def __call__(self, **args: Dict[str, Any]) -> NamedTuple:
        fields = [(name, type(val)) for name, val in args.items()]
        return NamedTuple('Outputs', fields)(**args)  # pylint: disable=not-callable



# Copyright 2021 Microsoft . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
from torch import nn
from towhee.models.utils.general_utils import to_2tuple

class PatchEmbed2D(nn.Module):
    """ 2D Image to Patch Embedding
    """
    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768, norm_layer=None, flatten=True):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        self.img_size = img_size
        self.patch_size = patch_size
        self.grid_size = (img_size[0] // patch_size[0], img_size[1] // patch_size[1])
        self.num_patches = self.grid_size[0] * self.grid_size[1]
        self.flatten = flatten

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)
        self.norm = norm_layer(embed_dim) if norm_layer else nn.Identity()

    def forward(self, x):
        _, _, h, w = x.shape
        assert h == self.img_size[0] and w == self.img_size[1], \
            f'Input image size ({h}*{w}) doesn\'t match model ({self.img_size[0]}*{self.img_size[1]}).'
        x = self.proj(x)
        if self.flatten:
            x = x.flatten(2).transpose(1, 2)  # BCHW -> BNC
        x = self.norm(x)
        return x





# original code from https://github.com/SwinTransformer/Video-Swin-Transformer
# modified by Zilliz.

import torch
from torch import nn
from torch.utils import checkpoint
import torch.nn.functional as F

from towhee.models.layers.window_attention3d import WindowAttention3D
from towhee.models.utils.window_partition import window_partition
from towhee.models.utils.window_reverse import window_reverse
from towhee.models.layers.mlp import Mlp
from towhee.models.utils.get_window_size import get_window_size
from towhee.models.layers.droppath import DropPath


class SwinTransformerBlock3D(nn.Module):
    """
    3D Swin Transformer Block.

    Args:
        dim (`int`):
            Number of input channels.
        num_heads (`int`):
            Number of attention heads.
        window_size (`tuple[int]`):
            Window size.
        shift_size (`tuple[int]`):
            Shift size for SW-MSA.
        mlp_ratio (`float`):
            Ratio of mlp hidden dim to embedding dim.
        qkv_bias (`bool`):
            If True, add a learnable bias to query, key, value. Default: True
        qk_scale (`float`):
            Override default qk scale of head_dim ** -0.5 if set.
        drop (`float`):
            Dropout rate. Default: 0.0
        attn_drop (`float`):
            Attention dropout rate. Default: 0.0
        drop_path(`float`):
            Stochastic depth rate. Default: 0.0
        act_layer (`nn.Module`):
            Activation layer. Default: nn.GELU
        norm_layer (`nn.Module`):
            Normalization layer.  Default: nn.LayerNorm
    """

    def __init__(self, dim, num_heads, window_size=(2, 7, 7), shift_size=(0, 0, 0),
                 mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0., drop_path=0.,
                 act_layer=nn.GELU, norm_layer=nn.LayerNorm, use_checkpoint=False):
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.window_size = window_size
        self.shift_size = shift_size
        self.mlp_ratio = mlp_ratio
        self.use_checkpoint = use_checkpoint

        assert 0 <= self.shift_size[0] < self.window_size[0], "shift_size must in 0-window_size"
        assert 0 <= self.shift_size[1] < self.window_size[1], "shift_size must in 0-window_size"
        assert 0 <= self.shift_size[2] < self.window_size[2], "shift_size must in 0-window_size"

        self.norm1 = norm_layer(dim)
        self.attn = WindowAttention3D(
            dim, window_size=self.window_size, num_heads=num_heads,
            qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)

        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

    def forward_part1(self, x, mask_matrix):
        b, d, h, w, c = x.shape
        window_size, shift_size = \
            get_window_size((d, h, w), self.window_size, self.shift_size)  # pylint: disable=unbalanced-tuple-unpacking

        x = self.norm1(x)
        # pad feature maps to multiples of window size
        pad_l = pad_t = pad_d0 = 0
        pad_d1 = (window_size[0] - d % window_size[0]) % window_size[0]
        pad_b = (window_size[1] - h % window_size[1]) % window_size[1]
        pad_r = (window_size[2] - w % window_size[2]) % window_size[2]
        x = F.pad(x, (0, 0, pad_l, pad_r, pad_t, pad_b, pad_d0, pad_d1))
        _, dp, hp, wp, _ = x.shape
        # cyclic shift
        if any(i > 0 for i in shift_size):
            shifted_x = torch.roll(x, shifts=(-shift_size[0], -shift_size[1], -shift_size[2]), dims=(1, 2, 3))
            attn_mask = mask_matrix
        else:
            shifted_x = x
            attn_mask = None
        # partition windows
        x_windows = window_partition(shifted_x, window_size)  # b*nW, Wd*Wh*Ww, c
        # W-MSA/SW-MSA
        attn_windows = self.attn(x_windows, mask=attn_mask)  # b*nW, Wd*Wh*Ww, c
        # merge windows
        attn_windows = attn_windows.view(-1, *(window_size + (c,)))
        shifted_x = window_reverse(attn_windows, window_size, b, dp, hp, wp)  # b d' h' w' c
        # reverse cyclic shift
        if any(i > 0 for i in shift_size):
            x = torch.roll(shifted_x, shifts=(shift_size[0], shift_size[1], shift_size[2]), dims=(1, 2, 3))
        else:
            x = shifted_x

        if pad_d1 > 0 or pad_r > 0 or pad_b > 0:
            x = x[:, :d, :h, :w, :].contiguous()
        return x

    def forward_part2(self, x):
        return self.drop_path(self.mlp(self.norm2(x)))

    def forward(self, x, mask_matrix):
        """ Forward function.
        Args:
            x (`tensor`):
                Input feature, tensor size (B, D, H, W, C).
            mask_matrix (`tuple[int]`):
                Attention mask for cyclic shift.
        """

        shortcut = x
        if self.use_checkpoint:
            x = checkpoint.checkpoint(self.forward_part1, x, mask_matrix)
        else:
            x = self.forward_part1(x, mask_matrix)
        x = shortcut + self.drop_path(x)

        if self.use_checkpoint:
            x = x + checkpoint.checkpoint(self.forward_part2, x)
        else:
            x = x + self.forward_part2(x)

        return x



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
from torch import nn

from towhee.models.utils.create_conv2d_pad import create_conv2d_pad
from towhee.models.layers.cond_conv2d import CondConv2d
from towhee.models.layers.mixed_conv2d import MixedConv2d

def create_conv2d(in_channels: int, out_channels: int, kernel_size: int, **kwargs) -> nn.Module:
    """ Select a 2d convolution implementation based on arguments
    Creates and returns one of torch.nn.Conv2d, Conv2dSame, MixedConv2d, or CondConv2d.
    Used extensively by EfficientNet, MobileNetv3 and related networks.
    """
    if isinstance(kernel_size, list):
        assert 'num_experts' not in kwargs  # MixNet + CondConv combo not supported currently
        assert 'groups' not in kwargs  # MixedConv groups are defined by kernel list
        # We're going to use only lists for defining the MixedConv2d kernel groups,
        # ints, tuples, other iterables will continue to pass to normal conv and specify h, w.
        m = MixedConv2d(in_channels, out_channels, kernel_size, **kwargs)
    else:
        depthwise = kwargs.pop('depthwise', False)
        # for DW out_channels must be multiple of in_channels as must have out_channels % groups == 0
        groups = in_channels if depthwise else kwargs.pop('groups', 1)
        if 'num_experts' in kwargs and kwargs['num_experts'] > 0:
            m = CondConv2d(in_channels, out_channels, kernel_size, groups=groups, **kwargs)
        else:
            m = create_conv2d_pad(in_channels, out_channels, kernel_size, groups=groups, **kwargs)
    return m



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from towhee.engine.operator_runner.runner_base import RunnerBase


class MapRunner(RunnerBase):
    """
    Map wrapper, one input one output.

    Wrapper will run in task executor in another thread or process.
    If run an op error, we should pass error info by an error handler.
    """



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import threading
from enum import Enum
from typing import List, Tuple, Any


from towhee.dataframe.array.array import Array
from towhee.dataframe._schema import _Schema
from towhee.types._frame import _Frame, FRAME
# from towhee.types import equivalents


class DataFrame:
    """
    A `DataFrame` is a collection of immutable, potentixally heterogeneous blogs of data.

    Args:
        columns (`list[Tuple[str, Any]]`)
            The list of the column names and their corresponding data types.
        name (`str`):
            Name of the dataframe; `DataFrame` names should be the same as its
            representation.
    """

    def __init__(
        self,
        name: str,
        columns: List[Tuple[str, str]],
    ):
        self._name = name
        self._sealed = False

        self._insert_cache_lock = threading.RLock()
        self._insert_cache = {}
        self._next_id = 0

        self._iterator_lock = threading.RLock()
        self._it_id = 0
        self._iterators = {}
        self._min_offset = 0

        self._block_lock = threading.RLock()
        self._map_blocked = {}
        self._window_start_blocked = {}
        self._window_end_blocked = {}

        self._data_lock = threading.RLock()
        self._len = 0
        self._data_as_list = []
        self._schema = _Schema()

        self._initialize_storage(columns)

    def _initialize_storage(self, columns):
        """Set the columns in the schema."""
        for name, col_type in columns:
            self._schema.add_col(name=name, col_type = col_type)
        frame = _Frame()
        self._schema.add_col(FRAME, self._class_type(frame))
        self._data_as_list = [Array(name=name) for name, _ in self._schema.cols]

    def __getitem__(self, key):
        """Get data at the passed in offset."""
        with self._data_lock:
            # access a row
            if isinstance(key, int):
                return tuple(self._data_as_list[i][key] for i in range(len(self._data_as_list)))
            # access a column
            elif isinstance(key, str):
                index = self._schema.col_index(key)
                return self._data_as_list[index]

    def __str__(self):
        """
        Simple to_string for printing and debugging dfs. Currently assumes that the data can be str()'ed.
        """
        ret = ''
        formater = ''
        columns = []
        with self._data_lock:
            for x in range(len(self._data_as_list)):
                columns.append(self._data_as_list[x].name)
                formater += '{' + str(x) + ':30}'
            ret += formater.format(*columns) + '\n'

            if self._min_offset != float('inf'):
                for x in range(self._min_offset, self._min_offset + self.__len__()):
                    values = []
                    for i in range(len(self._data_as_list)):
                        val = self._data_as_list[i][x]
                        values.append(str(val))
                    ret += formater.format(*values) + '\n'

            return ret

    def __len__(self):
        with self._data_lock:
            return self._len - self._min_offset

    @property
    def current_size(self):
        return self.__len__()

    @property
    def size(self):
        return self._len

    @property
    def name(self) -> str:
        return self._name

    @property
    def iterators(self) -> List[int]:
        return self._iterators

    @property
    def data(self) -> List[Array]:
        with self._data_lock:
            return self._data_as_list

    @property
    def columns(self) -> List[str]:
        return [x for x, _ in self._schema.cols]

    @property
    def types(self) -> List[Any]:
        return [y for _, y in self._schema.cols]

    @property
    def sealed(self) -> bool:
        return self._sealed

    def get_window(self, start: int, end: int, step: int, comparator: str, iter_id: int = False):
        """
        Window based data retrieval. Windows include everything up to but not including
        the cutoff.

        Args:
            start (`int`):
                The starting index to read data from.
            end (`int`):
                The current upper window limit.
            step (`int`):
                How far the window will slide for next call
            comparator ('timestamp' | 'row_id'):
                Which value is being used for window calculations.
            iter_id (`int`):
                Which iterator is reading the data.

        Returns:
            Responses:
                Different states of response codes depending on the dataframe state.

        Situations:

        Unsealed:
            [0, 1, 2, 3] window = (5, 10)
            1. Window start is greater than the latest value in df:
                a. block for value greater than window start

            [0, 1, 2, 3] window = (0, 10)
            2. Window end is greater than the latest value in df:
                a. block for value that is greater than window end

            [5, 6, 7, 8] window = (1, 4)
            3. Window end is smaller than all values in df
                a. return None and continue to next window

            [0, 1, 3, 4] window = (0, 3)
            4. Values exist between window start and window end:
                a. return values, clear data up to first value

        Sealed:
            [0, 1, 2, 3] window = (5, 10)
            1. Window start is greater than the latest value in df:
                a. Stop iteration

            [0, 1, 2, 3] window = (0, 10)
            2. Window end is greater than the latest value in df:
                a. return remaining window values and stop iteration

            [5, 6, 7, 8] window = (1, 4)
            3. Window end is smaller than all values in df
                a. return None and continue to next window

            [0, 1, 3, 4] window = (0, 3)
            4. Values exist between window start and window end:
                a. return values, clear data up to first value

        """

        # When the iterator is forcefully unblocked using unblock_iters().
        with self._iterator_lock:
            if iter_id is not False and self._iterators.get(iter_id) is None:
                return Responses.KILLED, None, None

        base_offset = self._iterators[iter_id]
        rets = []
        min_offset = None
        max_offset = base_offset
        next_offset = None
        next_window = start + step

        with self._data_lock:
            #  If the df is empty, we dont want to do anything but wait for a next value if unsealed.
            if self.__len__() == 0:
                if self.sealed:
                    return self._ret(Responses.EMPTY_SEALED, None, None, iter_id)
                else:
                    return self._ret(Responses.EMPTY, None, None, iter_id)
            #  If the first element in the df is larger than the window end we want to move to the next viable window
            if getattr(self._data_as_list[-1][base_offset], comparator) >= end:
                goal = getattr(self._data_as_list[-1][base_offset], comparator)
                new_start = start + step * ((goal - start)//step)
                new_end = (new_start - start) + end
                if new_end <= goal:
                    new_start += step
                    new_end += step
                return self._ret(Responses.OLD_WINDOW, None, (new_start, (new_start - start) + end), iter_id)

            #  If the last element of the df is smaller than the start of the window, proceed to next windows.
            elif getattr(self._data_as_list[-1][-1], comparator) < start:
                if self.sealed:
                    return self._ret(Responses.FUTURE_WINDOW_SEALED, None, None, iter_id)
                else:
                    return self._ret(Responses.FUTURE_WINDOW, None, self._len, iter_id)

            # iterating through values to see which ones fall in the window. min offset = first value, max offset = last value
            for i in range(base_offset, self._len):
                if getattr(self._data_as_list[-1][i], comparator) >= start and getattr(self._data_as_list[-1][i], comparator) < end:
                    if min_offset is None:
                        min_offset = i
                    if next_offset is None:
                        if getattr(self._data_as_list[-1][i], comparator) >= next_window:
                            next_offset = i
                    max_offset = i
                    rets.append(self.__getitem__(max_offset))
                elif getattr(self._data_as_list[-1][i], comparator) >= end:
                    if next_offset is None:
                        next_offset = i
                    break

            # If the max offset is the length of the df
            if max_offset == self._len - 1 and self.sealed:
                # Need to check if there is a next window that fits, if so, continue
                if next_offset:
                    return self._ret(Responses.APPROVED_CONTINUE, rets, next_offset, iter_id)
                # If not, return whats left and have iterator close.
                else:
                    return self._ret(Responses.APPROVED_DONE, rets, max_offset, iter_id)
            # If the last value that fits in window is also the last value of df, cant close
            # window yet as we dont know if the next value fits in the same window.
            elif max_offset == self._len - 1 and not self.sealed:
                return self._ret(Responses.WINDOW_NOT_DONE, None, None, iter_id)
            else:
                return self._ret(Responses.APPROVED_CONTINUE, rets, next_offset, iter_id)


    def get(self, offset: int, count: int = 1, iter_id: int = False):
        """
        Get data from dataframe based on offset and how many values requested.

        Args:
            offset (`int`):
                The starting index to read data from.
            count (`int`):
                How many values to retrieve.
            iter_id (`int`):
                Which iterator is reading the data.

        Returns:
            Responses:
                Different states of response codes depending on the dataframe state.
        """

        # When the iterator is forcefully unblocked using unblock_iters().
        with self._iterator_lock:
            if iter_id is not False and self._iterators.get(iter_id) is None:
                return self._ret(Responses.KILLED, None, None, iter_id)

            elif offset < self._min_offset:
                # Up to iterators to decide what to do if the value they want is GCed.
                return self._ret(Responses.INDEX_GC, None, None, iter_id)

        with self._data_lock:
            # The batch of data is available and there is more available.
            if offset + count <= self._len:
                return self._ret(Responses.APPROVED_CONTINUE, [self.__getitem__(x) for x in range(offset, offset + count)], None, iter_id)

            elif self._sealed:
                # If no more data is left and trying to get more, iterator will stop and no data returned.
                if offset >= self._len:
                    return self._ret(Responses.INDEX_OOB_SEALED, None, None, iter_id)

                # If dataframe is sealed, remaining values will be returned and allow for next step
                else:
                    return self._ret(Responses.APPROVED_CONTINUE, [self.__getitem__(x) for x in range(offset, self._len)], None, iter_id)

            # If the requested offset is out of bounds but dataframe is still writing data.
            elif offset + count > self._len:
                return self._ret(Responses.INDEX_OOB_UNSEALED, None, None, iter_id)

            # Something broke.
            else:
                return self._ret(Responses.UNKOWN_ERROR, None, None, iter_id)

    def _ret(self, code, ret, offset, provide_code):
        if  provide_code is not False:
            return code, ret, offset
        else:
            return ret


    def put(self, item) -> None:
        """
        Append a new value to the dataframe.

        Args:
            item (dict | list | Array | tuple):
                The row to insert.

        """
        assert not self._sealed, f'DataFrame {self._name} is already sealed, can not put data'
        assert isinstance(item, (tuple, dict, list)), f'Dataframe input must be of type (tuple, dict, list), not {type(item)}'
        with  self._data_lock:
            start_len = self._len
            if not isinstance(item, list):
                item = [item]
            for x in item:
                self._put(x)

             # Clear cache after inserting.
            self._put_cached()

            frame = None
            # Only check blocked iterators if new value added.
            if (self._len - start_len) > 0 and self.current_size > 0:
                frame = self._data_as_list[-1][-1]
                cur_len = self._len

        if frame is not None:
            self._check_blocked(frame, cur_len)

    def _check_blocked(self, frame, cur_len):
        """Unblock iterators if their criteria has been met."""
        with self._block_lock:
            if len(self._map_blocked) > 0:
                rem = []
                for i, id_events in self._map_blocked.items():
                    if i <= cur_len:
                        for _, event in id_events:
                            event.set()
                        rem.append(i)
                for key in rem:
                    del self._map_blocked[key]
            if len(self._window_start_blocked) > 0:
                rem = []
                for cutoff, id_events in self._window_start_blocked.items():
                    if cutoff[1] <= getattr(frame, cutoff[0]):
                        for _, event in id_events:
                            event.set()
                        rem.append(cutoff)
                for key in rem:
                    del self._window_start_blocked[key]
            if len(self._window_end_blocked) > 0:
                rem = []
                for cutoff, id_events in self._window_end_blocked.items():
                    if cutoff[1] < getattr(frame, cutoff[0]):
                        for _, event in id_events:
                            event.set()
                        rem.append(cutoff)
                for key in rem:
                    del self._window_end_blocked[key]

    def _put(self, item):
        """Method that decides which format the data is in and calls put."""
        if isinstance(item, dict):
            return self._put_dict(item)
        elif isinstance(item, tuple):
            return self._put_tuple(item)
        else:
            raise ValueError('Input data is of wrong format.')

    def _put_cached(self):
        """Empty cache values into the data list when the index is reached."""
        if len(self._insert_cache) != 0:
            with self._insert_cache_lock:
                item = self._insert_cache.pop(self._next_id, None)
                while item is not None:
                    # If the row is a frame, skip it but increment next_id
                    if item == -1:
                        self._next_id += 1
                    else:
                        self._put(item)
                    item = self._insert_cache.pop(self._next_id, None)

    def _put_tuple(self, item: tuple):
        """Appending a tuple to the dataframe."""
        if not isinstance(item[-1], _Frame):
            item = item + (_Frame(row_id=self._len), )
        else:
            # If row is empty, append to the insert cache for skipping.
            if item[-1].empty:
                with self._insert_cache_lock:
                    self._insert_cache[item[-1].prev_id] = -1
                    return -1
            # if prev_id exists and its not equal to next_id, cache it
            elif item[-1].prev_id != -1 and item[-1].prev_id != self._next_id:
                with self._insert_cache_lock:
                    self._insert_cache[item[-1].prev_id] = item
                    return -2
            item[-1].row_id = self._len

        # TODO: Figure out the situation on type checking
        # for i, x in enumerate(item):
            # print(equivalents.get(self._class_type(x), self._class_type(x)) , self._schema.col_type(i), flush=True)
            # assert equivalents.get(self._class_type(x), self._class_type(x)) == self._schema.col_type(i)

        for i, x in enumerate(item):
            self._data_as_list[i].put(x)

        self._len += 1
        self._next_id += 1

    def _put_dict(self, item: dict):
        """Appending a dict to the dataframe."""
        if item.get(FRAME, None) is None:
            item[FRAME] = _Frame(self._len)
        else:
            # If row is empty, append to the insert cache for skipping.
            if item[FRAME].empty:
                with self._insert_cache_lock:
                    self._insert_cache[item[-1].prev_id] = -1
                    return -1
            # if prev_id exists and its not equal to next_id, cache it
            elif item[FRAME].prev_id != -1 and item[FRAME].prev_id != self._next_id:
                with self._insert_cache_lock:
                    self._insert_cache[item[FRAME].prev_id] = item
                    return -2
            item[FRAME].row_id = self._len

        # I believe its faster to loop through and check than list comp
        # TODO: Add type checking
        # for key, val in item.items():
            # assert equivalents.get(self._class_type(val), self._class_type(val)) == self._schema.col_type(self._schema.col_index(key))

        for key, val in item.items():
            self._data_as_list[self._schema.col_index(key)].put(val)

        self._len += 1
        self._next_id += 1

    def seal(self):
        """
        Function to seal the dataframe. A sealed dataframe will not accept anymore data and
        the act of sealing the dataframe will trigger blocked iterators to pull the remaining data.
        """
        with self._data_lock:
            self._sealed = True

        # Release all blocked iters.
        with self._block_lock:
            for _, id_event in self._map_blocked.items():
                for _, event in id_event:
                    event.set()
            self._map_blocked.clear()
            for _, id_event in self._window_start_blocked.items():
                for _, event in id_event:
                    event.set()
            self._window_start_blocked.clear()
            for _, id_event in self._window_end_blocked.items():
                for _, event in id_event:
                    event.set()
            self._window_end_blocked.clear()

    def gc(self):
        self.gc_data()
        self.gc_blocked()

    def gc_data(self):
        """Garbage collection function to trigger towhee.Array GC."""
        with self._iterator_lock:
            vals = [value for _, value in self._iterators.items()]
            self._min_offset = min(vals + [self._len])
        with self._data_lock:
            for x in self._data_as_list:
                x.gc(self._min_offset)

    def gc_blocked(self):
        """Garbage collection for blocked iterator list."""
        with self._block_lock:
            for key, val in list(self._map_blocked.items()):
                if len(val) == 0:
                    del self._map_blocked[key]
            for key, val in list(self._window_start_blocked.items()):
                if len(val) == 0:
                    del self._window_start_blocked[key]
            for key, val in list(self._window_end_blocked.items()):
                if len(val) == 0:
                    del self._window_end_blocked[key]

    def register_iter(self):
        """
        Registering an iter allows the df to keep track of where each iterator is in order
        to coordinate garbage collection and other processes.

        Returns:
            An iterator ID that is used for a majority of iterator-dataframe interaction.
        """
        with self._iterator_lock:
            self._it_id += 1
            self._iterators[self._it_id] = 0
            return self._it_id

    def remove_iter(self, iter_id: int):
        """
        Removing the iterator from the df. Allows for more garbage collecting.
        Args:
            iter_id (`int`):
                The iterator's id.
        """
        self.unblock_iter(iter_id)
        try:
            with self._iterator_lock:
                if iter_id in self._iterators:
                    del self._iterators[iter_id]
        except KeyError:
            pass

        # self.gc_data()

    def ack(self, offset: int, iter_id: int):
        """
        An acknowledgement (ack) will notice the `DataFrame`s that the rows already
        iterated over are no longer used, and can be garbage collected. Up to this index
        but not including.

        Args:
            offset (`int`):
                The latest accepted offset.
            iter_id (`int`):
                The iterator id to set offset for.
        """
        with self._iterator_lock:
            if self._iterators[iter_id] <= (offset):
                self._iterators[iter_id] = offset
        # TODO: Figure out better algorithm for gc'ing
        # self.gc_data()

    def notify_map_block(self, event: threading.Event, offset: int, count: int, iter_id: int):
        """
        Used by map based iterators to notify the dataframe that it is waiting on a value.

        Args:
            event (`threading.Event`):
                The event to trigger once the value being waited on is available.
            offset (`int`):
                The offset that is being waited on.
            count (`int):
                How many values are being waited on.
            iter_id (`int`):
                The iterator id of the blocked iterator.
        """
        with self._block_lock:
            index = offset + count
            if self._map_blocked.get(index) is None:
                self._map_blocked[index] = []
            self._map_blocked[index].append((iter_id, event))

    def notify_window_block(self, event: threading.Event, edge: str, cutoff: Tuple[str, int], iter_id: int):
        """
        Used by window based iterators to notify the dataframe that it is waiting on a value.

        Args:
            iter_id (`int`):
                The iterator id of the blocked iterator.
            event (`threading.Event`):
                The event to trigger once the value being waited on is available.
            edge ('start', 'end'):
                Waiting for a value >= than start, or waiting for a value greater then end.
            cutoff ("timestamp" | "row_id", `int`):
                The window cutoff that is is being waited on.
        """

        with self._block_lock:
            if edge == 'start':
                if self._window_start_blocked.get(cutoff) is None:
                    self._window_start_blocked[cutoff] = []
                self._window_start_blocked[cutoff].append((iter_id, event))

            elif edge == 'end':
                if self._window_end_blocked.get(cutoff) is None:
                    self._window_end_blocked[cutoff] = []
                self._window_end_blocked[cutoff].append((iter_id, event))

    def unblock_all(self):
        """Forceful unblocking/killing of all blocked iter."""
        with self._block_lock:
            for _, id_event in self._map_blocked.items():
                for it_id, event in id_event:
                    del self._iterators[it_id]
                    event.set()

            self._map_blocked.clear()

            for _, id_event in self._window_start_blocked.items():
                for it_id, event in id_event:
                    del self._iterators[it_id]
                    event.set()

            self._window_start_blocked.clear()

            for _, id_event in self._window_end_blocked.items():
                for it_id, event in id_event:
                    del self._iterators[it_id]
                    event.set()

            self._window_end_blocked.clear()

    def unblock_iter(self, iter_id: int):
        """
        Forceful unblocking/killing of the selected iter.

        Args:
            iter_id (`int`):
                The id of the iterator being unblocked.
        """
        with self._block_lock:
            index = None
            for _, id_event in self._map_blocked.items():
                for i, (ids, event) in enumerate(id_event):
                    if ids == iter_id:
                        del self._iterators[iter_id]
                        event.set()
                        index = i
                        break
                if index is not None:
                    del id_event[index]
                    # self.gc_blocked()
                    return

            for _, id_event in self._window_start_blocked.items():
                for i, (ids, event) in enumerate(id_event):
                    if ids == iter_id:
                        del self._iterators[iter_id]
                        event.set()
                        index = i
                        break
                if index is not None:
                    del id_event[index]
                    # self.gc_blocked()
                    return

            for _, id_event in self._window_end_blocked.items():
                for i, (ids, event) in enumerate(id_event):
                    if ids == iter_id:
                        del self._iterators[iter_id]
                        event.set()
                        index = i
                        break
                if index is not None:
                    del id_event[index]
                    # self.gc_blocked()
                    return

    def _class_type(self, o):
        """Iternal function to find the full module path of object."""
        module = o.__class__.__module__
        if module is None or module == str.__class__.__module__:
            return o.__class__.__name__
        return module + '.' + o.__class__.__name__


class Responses(Enum):
    """Response Codes between DF and Iterators."""
    INDEX_GC = 1
    INDEX_OOB_UNSEALED = 2
    INDEX_OOB_SEALED = 3
    APPROVED_CONTINUE = 4
    APPROVED_DONE = 5
    KILLED = 6
    UNKOWN_ERROR = 7
    WINDOW_NOT_DONE = 8
    WINDOW_PASSED = 9
    FUTURE_WINDOW = 10
    OLD_WINDOW = 11
    FUTURE_WINDOW_SEALED = 12
    EMPTY = 13
    EMPTY_SEALED = 14



# original code from https://github.com/SwinTransformer/Video-Swin-Transformer
# modified by Zilliz.

from torch import nn
import torch.nn.functional as F


class PatchEmbed3D(nn.Module):
    """
    3D patch embedding for video.

    Args:
        patch_size (`int`):
            Patch token size.
        c (`int`):
            Number of input video channels..
        embed_dim (`int`):
            Number of linear projection output channels.
        norm_layer (`nn.Module`):
            Normalization layer.
    """
    def __init__(self, patch_size=(2, 4, 4), c=3, embed_dim=96, norm_layer=None):
        super().__init__()
        self.patch_size = patch_size
        self.c = c
        self.embed_dim = embed_dim

        self.proj = nn.Conv3d(c, embed_dim, kernel_size=patch_size, stride=patch_size)
        if norm_layer is not None:
            self.norm = norm_layer(embed_dim)
        else:
            self.norm = None

    def forward(self, x):
        # do padding
        _, _, d, h, w = x.size()
        if w % self.patch_size[2] != 0:
            x = F.pad(x, (0, self.patch_size[2] - w % self.patch_size[2]))
        if h % self.patch_size[1] != 0:
            x = F.pad(x, (0, 0, 0, self.patch_size[1] - h % self.patch_size[1]))
        if d % self.patch_size[0] != 0:
            x = F.pad(x, (0, 0, 0, 0, 0, self.patch_size[0] - d % self.patch_size[0]))

        x = self.proj(x)  # b c d h w
        if self.norm is not None:
            d, h, w = x.size(2), x.size(3), x.size(4)
            x = x.flatten(2).transpose(1, 2)
            x = self.norm(x)
            x = x.transpose(1, 2).view(-1, self.embed_dim, d, h, w)

        return x



# original code from https://github.com/krasserm/perceiver-io
# modified by Zilliz


import os

try:
    import fairscale
except ModuleNotFoundError:
    os.system("pip install fairscale")
from fairscale.nn import checkpoint_wrapper

from towhee.models.perceiver.Residual import Residual
from towhee.models.perceiver.Sequential import Sequential
from towhee.models.perceiver.mlp import mlp
from towhee.models.perceiver.CrossAttention import CrossAttention


def cross_attention_layer(
    num_q_channels: int, num_kv_channels: int, num_heads: int, dropout: float, activation_checkpoint: bool = False
):
    """
    Cross attention block for Perceiver https://arxiv.org/pdf/2103.03206.pdf.

    Args:
        num_q_channels (`int`):
            Number of q channels.
        num_kv_channels (`int`):
            Number of k or v channels. k has the same channels as v.
        num_heads (`int`):
            Number of parallel attention heads.
        dropout (`nn.Module`):
            Dropout probability.
        activation_checkpoint (`bool`):
            Use activation checkpointing.

    Return (`nn.Module`):
        Configured cross attention layer.
    """
    layer = Sequential(
        Residual(CrossAttention(num_q_channels, num_kv_channels, num_heads, dropout), dropout),
        Residual(mlp(num_q_channels), dropout),
    )
    return layer if not activation_checkpoint else checkpoint_wrapper(layer)



# Copyright 2021 Ross Wightman . All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code is modified by Zilliz.
import math
from typing import List, Tuple
import torch
import torch.nn.functional as F

# Calculate symmetric padding for a convolution
def get_padding(kernel_size: int, stride: int = 1, dilation: int = 1, **_) -> int:
    """
    Calculate symmetric padding for a convolution
    Args:
        kernel_size(`Int`):
            Convolution kernel size.
        stride(`Int`):
            Convolution stride parameter.
        dilation(`Int`):
            Convolution dilation parameter.
    Returns:
        (`Int`)
            Padding size to keep.
    """
    padding = ((stride - 1) + dilation * (kernel_size - 1)) // 2
    return padding

def get_same_padding(x: int, k: int, s: int, d: int) -> int:
    """
    Calculate asymmetric TensorFlow-like 'SAME' padding for a convolution
    Args:
        x(`Int`):
            Input tensor shape.
        k(`Int`):
            Convolution kernel size.
        s(`Int`):
            Convolution stride parameter.
        d(`Int`):
            Convolution dilation parameter.
    Returns:
        (`Int`):
            Padding value for 'SAME' padding.
    """
    return max((math.ceil(x / s) - 1) * s + (k - 1) * d + 1 - x, 0)

def is_static_pad(kernel_size: int, stride: int = 1, dilation: int = 1, **_) -> bool:
    """
    Can SAME padding for given args be done statically?
    Args:
        kernel_size(`Int`):
            Convolution kernel size.
        stride(`Int`):
            Convolution stride parameter.
        dilation(`Int`):
            Convolution dilation parameter.
    Returns:
        (`Bool`): whether SAME padding can be done statically.
    """
    return stride == 1 and (dilation * (kernel_size - 1)) % 2 == 0

def pad_same(x: torch.Tensor, k: List[int], s: List[int], d: List[int] = (1, 1), value: float = 0) -> torch.Tensor:
    """
    Dynamically pad input x with 'SAME' padding for conv with specified args
    Args:
        x(`torch.Tensor`):
            Input tensor.
        k(`List[Int]`):
            Convolution kernel sizes.
        s(`List[Int]`):
            Convolution stride parameters.
        d(`List[Int]`):
            Convolution dilation parameter.
        value(`Float`):
            Value for padding.
    Returns:
        (`torch.Tensor`):
            Output Tensor for conv with 'SAME' padding.
    """
    ih, iw = x.size()[-2:]
    pad_h, pad_w = get_same_padding(ih, k[0], s[0], d[0]), get_same_padding(iw, k[1], s[1], d[1])
    if pad_h > 0 or pad_w > 0:
        x = F.pad(x, [pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2], value=value)
    return x

def get_padding_value(padding: str, kernel_size: int, **kwargs) -> Tuple[int, bool]:
    """
    Args:
        padding(`Str`):
            Padding type, 'same' or 'valid'.
        kernel_size(`Int`):
            Convolution kernel size.
    Returns:
        (`Int`): padding shape.
        (`Bool`): dynamically padding.
    """
    dynamic = False
    if isinstance(padding, str):
        # for any string padding, the padding will be calculated for you, one of three ways
        padding = padding.lower()
        if padding == 'same':
            # TF compatible 'SAME' padding, has a performance and GPU memory allocation impact
            if is_static_pad(kernel_size, **kwargs):
                # static case, no extra overhead
                padding = get_padding(kernel_size, **kwargs)
            else:
                # dynamic 'SAME' padding, has runtime/GPU memory overhead
                padding = 0
                dynamic = True
        elif padding == 'valid':
            # 'VALID' padding, same as padding=0
            padding = 0
        else:
            # Default to PyTorch style 'same'-ish symmetric padding
            padding = get_padding(kernel_size, **kwargs)
    return padding, dynamic



# Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from typing import Tuple

from towhee.hparam import param_scope


class Collector:
    """
    Collector class for metric
    """

    # pylint: disable=dangerous-default-value
    def __init__(self,
                 metrics: list = None,
                 labels: dict = None,
                 scores: dict = None):
        self.metrics = metrics if metrics is not None else []
        self.scores = scores if scores is not None else {}
        self.labels = labels if labels is not None else {}

    def add_metrics(self, value: str):
        self.metrics.append(value)

    def add_scores(self, value: dict):
        self.scores.update(value)

    def add_labels(self, value: dict):
        self.labels.update(value)


def encode_fig_img(mat, size=300):
    # pylint: disable=import-outside-toplevel
    from towhee.utils.sklearn_utils import ConfusionMatrixDisplay
    from towhee.utils.matplotlib_utils import matplotlib as mpl
    import base64
    import io
    mpl.use('Agg')  # Prevent showing stuff
    cm = mat.astype('float') / mat.sum(axis=1)[:, np.newaxis]  # normalize
    fig = ConfusionMatrixDisplay(cm)
    fig.plot(cmap='GnBu')
    buf = io.BytesIO()
    fig.figure_.savefig(buf, format='jpg')
    buf.seek(0)
    buf = buf.read()
    src = 'src="data:image/jpeg;base64,' + base64.b64encode(buf).decode() + '" '
    w = 'width = "' + str(size) + 'px" '
    h = 'height = "' + str(size) + 'px" '
    return '<img ' + src + w + h + '>'


def get_scores_dict(collector: Collector):
    scores_dict = {}
    for metric in collector.metrics:
        scores_dict[metric] = []
        for model in collector.scores:
            scores_dict[metric].append(collector.scores[model][metric])

    return scores_dict


def _evaluate_callback(self):

    def wrapper(_: str, index: Tuple[str], *arg, **kws):
        # pylint: disable=import-outside-toplevel
        # pylint: disable=unused-argument
        actual, predicted = index
        name = None
        if 'name' in kws:
            name = kws['name']
        elif arg:
            name, = arg
        self.collector.add_labels(
            {name: {
                'actual': actual,
                'predicted': predicted
            }})
        score = {name: {}}
        actual_list = []
        predicted_list = []
        for x in self:
            actual_list.append(getattr(x, actual))
            predicted_list.append(getattr(x, predicted))

        from towhee.utils import sklearn_utils
        for metric_type in self.collector.metrics:
            if metric_type == 'accuracy':
                re = sklearn_utils.accuracy_score(actual_list, predicted_list)
            elif metric_type == 'recall':
                re = sklearn_utils.recall_score(actual_list,
                                                predicted_list,
                                                average='weighted')
            elif metric_type == 'confusion_matrix':
                re = sklearn_utils.confusion_matrix(actual_list,
                                                    predicted_list)
            score[name].update({metric_type: re})
        self.collector.add_scores(score)
        return self

    return wrapper


class MetricMixin:
    """
    Mixin for metric
    """

    # pylint: disable=import-outside-toplevel
    def __init__(self):
        super().__init__()
        self.collector = Collector()
        self.evaluate = param_scope().callholder(_evaluate_callback(self))

    def with_metrics(self, metric_types: list = None):
        self.collector.metrics = metric_types
        return self

    def report(self):
        """
        report the metric scores

        Examples:

        >>> from towhee import DataCollection
        >>> from towhee import Entity
        >>> dc1 = DataCollection([Entity(a=a, b=b, c=c) for a, b, c in zip([0,1,1,0,0], [0,1,1,1,0], [0,1,1,0,0])])
        >>> dc1.with_metrics(['accuracy', 'recall']).evaluate['a', 'c'](name='lr').evaluate['a', 'b'](name='rf').report()
            accuracy  recall
        lr       1.0     1.0
        rf       0.8     0.8
        {'lr': {'accuracy': 1.0, 'recall': 1.0}, 'rf': {'accuracy': 0.8, 'recall': 0.8}}

        """
        from towhee.utils.ipython_utils import HTML, display
        from towhee.utils.pandas_utils import pandas as pd
        scores_dict = get_scores_dict(self.collector)
        df = pd.DataFrame(data=scores_dict, index=list(self.collector.scores.keys()))

        if 'confusion_matrix' in self.collector.metrics:
            display(HTML(df.to_html(formatters={'confusion_matrix': encode_fig_img}, escape=False)))
        else:
            display(df)
        return self.collector.scores


if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=False, optionflags=doctest.ELLIPSIS)



# VGGish is an auido embedding model developed by Tensorflow:
# https://github.com/tensorflow/models/tree/master/research/audioset/vggish
#
# Pytorch implementation is adapted from: https://github.com/harritaylor/torch-vggish
#
# All modifications are made by / Copyright 2021 Zilliz. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from torch import nn


class VGG(nn.Module):
    """
    PyTorch model class
    """
    def __init__(self):
        super().__init__()
        self.features = nn.Sequential(
            nn.Conv2d(1, 64, 3, 1, 1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(2, 2),
            nn.Conv2d(64, 128, 3, 1, 1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(2, 2),
            nn.Conv2d(128, 256, 3, 1, 1),
            nn.ReLU(inplace=True),
            nn.Conv2d(256, 256, 3, 1, 1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(2, 2),
            nn.Conv2d(256, 512, 3, 1, 1),
            nn.ReLU(inplace=True),
            nn.Conv2d(512, 512, 3, 1, 1),
            nn.ReLU(inplace=True),
            nn.MaxPool2d(2, 2))
        self.embeddings = nn.Sequential(
            nn.Linear(512 * 24, 4096),
            nn.ReLU(inplace=True),
            nn.Linear(4096, 4096),
            nn.ReLU(inplace=True),
            nn.Linear(4096, 128),
            #nn.ReLU(inplace=True)
        )

    def forward(self, x):
        x = self.features(x).permute(0, 2, 3, 1).contiguous()
        x = x.view(x.size(0), -1)
        x = self.embeddings(x)
        return x
